from pyxnat import Interface
from requests.exceptions import ConnectionError
import os.path as op
from nose import SkipTest
from functools import wraps
from nose.plugins.attrib import attr


fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')


def docker_available(func=None):
    '''Skip test completely if no Docker-based XNAT instance available
    '''

    def check_and_raise():
        if 'setup_docker_xnat' in func.__name__:
            print('Initializing XNAT.')
            return
        fp = op.abspath('.xnat.cfg')
        print(fp, op.isfile(fp))

        x = Interface(config=op.abspath('.xnat.cfg'))

        try:
            x.head('')
            list(x.select.projects())
            print('Docker instance found.')
        except (ConnectionError, KeyError):
            print('Skipping it.')
            raise SkipTest('Docker-based XNAT instance unavailable')

    if func:
        @wraps(func)
        @attr('network')
        @attr('docker_available')
        def newfunc(*args, **kwargs):
            check_and_raise()
            return func(*args, **kwargs)
        return newfunc
    else:
        check_and_raise()

@docker_available
def test_001_setup_docker_xnat():
    import os
    import pyxnat

    x = pyxnat.Interface(config='.xnat.cfg')

    try:
        cmd = 'curl --cookie-jar /tmp/cookie --header "Content-Type: application/x-www-form-urlencoded" '\
            '--request POST '\
            '--data "username={user}&password={password}&login=&XNAT_CSRF=" '\
            'http://localhost/login'.format(user=x._user, password=x._pwd)
        print(cmd)
        os.system(cmd)

        cmd = 'curl --cookie /tmp/cookie --header "Content-Type: application/json" '\
            '--request POST '\
            '--data \'%s\' '\
            'http://localhost/xapi/siteConfig'

        cmd2 = cmd % '{"siteId": "XNAT", "siteUrl": "http://localhost", "adminEmail": "fake@fake.fake"}'
        print(cmd2)
        os.system(cmd2)

        cmd2 = cmd % '{"archivePath":"/data/xnat/archive","prearchivePath":"/data/xnat/prearchive","cachePath":"/data/xnat/cache","buildPath":"/data/xnat/build","ftpPath":"/data/xnat/ftp","pipelinePath":"/data/xnat/pipeline","inboxPath":"/data/xnat/inbox"}'
        print(cmd2)
        os.system(cmd2)

        cmd2 = cmd % '{"requireLogin":true,"userRegistration":false,"enableCsrfToken":true}'
        print(cmd2)
        os.system(cmd2)

        cmd2 = cmd % '{"initialized":true}'
        print(cmd2)
        os.system(cmd2)


        p = x.select.project('nosetests')
        p.create()
        for i in range(3, 10):
            p = x.select.project('nosetests%s' % i)
            p.create()

        uri = '/data/projects/nosetests'
        options = {'alias': 'nosetests2'}
        data = x.put(uri, params=options).text

    except Exception:
        print('Skipping initialization.')
        raise SkipTest('Docker-based XNAT initialization failed.')



@docker_available
def test_001_setup_docker_xnat():
    import os
    import pyxnat

    x = pyxnat.Interface(config='.xnat.cfg')

    cmd = 'curl --cookie-jar /tmp/cookie --header "Content-Type: application/x-www-form-urlencoded" '\
        '--request POST '\
        '--data "username={user}&password={password}&login=&XNAT_CSRF=" '\
        'http://localhost/login'.format(user=x._user, password=x._pwd)
    print(cmd)
    os.system(cmd)

    cmd = 'curl --cookie /tmp/cookie --header "Content-Type: application/json" '\
        '--request POST '\
        '--data \'%s\' '\
        'http://localhost/xapi/siteConfig'

    cmd2 = cmd % '{"siteId": "XNAT", "siteUrl": "http://localhost", "adminEmail": "fake@fake.fake"}'
    print(cmd2)
    os.system(cmd2)

    cmd2 = cmd % '{"archivePath":"/data/xnat/archive","prearchivePath":"/data/xnat/prearchive","cachePath":"/data/xnat/cache","buildPath":"/data/xnat/build","ftpPath":"/data/xnat/ftp","pipelinePath":"/data/xnat/pipeline","inboxPath":"/data/xnat/inbox"}'
    print(cmd2)
    os.system(cmd2)

    cmd2 = cmd % '{"requireLogin":true,"userRegistration":false,"enableCsrfToken":true}'
    print(cmd2)
    os.system(cmd2)

    cmd2 = cmd % '{"initialized":true}'
    print(cmd2)
    os.system(cmd2)


    p = x.select.project('nosetests')
    p.create()
    for i in range(3, 10):
        p = x.select.project('nosetests%s' % i)
        p.create()

    uri = '/data/projects/nosetests'
    options = {'alias': 'nosetests2'}
    data = x.put(uri, params=options).text


@docker_available
def test_users():
    x = Interface(config='.xnat.cfg')
    assert isinstance(x.manage.users(), list)


@docker_available
def test_user_firstname():
    x = Interface(config='.xnat.cfg')
    assert x.manage.users.firstname('admin') == 'Admin'


@docker_available
def test_user_lastname():
    x = Interface(config='.xnat.cfg')
    assert x.manage.users.lastname('admin') == 'Admin'


@docker_available
def test_user_email():
    x = Interface(config='.xnat.cfg')
    assert x.manage.users.email('admin') == \
        'fake@fake.fake'


@docker_available
def test_user_id():
    x = Interface(config='.xnat.cfg')
    assert x.manage.users.id('admin') == '1'


@docker_available
def test_add_remove_user():
    x = Interface(config='.xnat.cfg')
    x.select.project('nosetests5').remove_user('admin')
    x.select.project('nosetests5').add_user('admin', 'collaborator')
    assert 'admin' in x.select.project('nosetests5').collaborators()
    x.select.project('nosetests5').remove_user('admin')
    assert 'admin' not in x.select.project('nosetests5').collaborators()
    x.select.project('nosetests5').add_user('admin', 'owner')


@docker_available
def test_project_accessibility():
    x = Interface(config='.xnat.cfg')
    print(x.select.project('nosetests5').accessibility())
    assert x.select.project('nosetests5').accessibility() in \
           [b'public', b'protected', b'private']
    x.select.project('nosetests5').set_accessibility('private')
    assert x.select.project('nosetests5').accessibility() == b'private'
    x.select.project('nosetests5').set_accessibility('protected')
    assert x.select.project('nosetests5').accessibility() == b'protected'


@docker_available
def test_create_xml():
    x = Interface(config='.xnat.cfg')
    _modulepath = op.dirname(op.abspath(__file__))
    p = x.select.project('nosetests')
    s = p.subject('10001')
    e = s.experiment('10001_MR')
    fp = op.join(_modulepath, 'sess.xml')
    assert(op.isfile(fp))
    e.create(xml=fp)
    xml = e.get().decode()
    assert('Alomar' in xml)


def test_project_users():
    x = Interface(config=fp)
    assert isinstance(x.select.project('nosetests5').users(), list)


def test_project_owners():
    x = Interface(config=fp)
    assert isinstance(x.select.project('nosetests5').owners(), list)


def test_project_members():
    x = Interface(config=fp)
    assert isinstance(x.select.project('nosetests5').members(), list)


def test_project_collaborators():
    x = Interface(config=fp)
    assert isinstance(x.select.project('nosetests5').collaborators(),
                      list)


def test_project_user_role():
    x = Interface(config=fp)
    assert x.select.project('nosetests5').user_role('nosetests') == 'owner'



from .schema import rest_translation
# from .schema import resources_types


def translate_uri(uri):
    segs = uri.split('/')
    for key in rest_translation.keys():
        if key in segs[-2:]:
            uri = uri.replace(key, rest_translation[key])

    return uri


def inv_translate_uri(uri):
    inv_table = dict(zip(rest_translation.values(), rest_translation.keys()))

    for key in inv_table.keys():
        uri = uri.replace('/%s' % key, '/%s' % inv_table[key])

    return uri


def join_uri(uri, *segments):
    part1 = [seg.lstrip('/') for seg in segments]
    return '/'.join(uri.split('/') + part1).rstrip('/')


def uri_last(uri):
    # return uri.split(uri_parent(uri))[1].strip('/')
    return uri.split('/')[-1]


def uri_nextlast(uri):
    # return uri_last(uri.split(uri_last(uri))[0].strip('/'))
    # support files in a hierarchy
    if '/files/' in uri:
        return 'files'
    return uri.split('/')[-2]


def uri_parent(uri):
    # parent = uri

    # if not os.path.split(uri)[1] in resources_types:
    #     while os.path.split(parent)[1] not in resources_types:
    #         parent = os.path.split(parent)[0]

    #     return parent

    # support files in a hierarchy by stripping all but one level
    files_index = uri.find('/files/')
    if files_index >= 0:
        uri = uri[:7+files_index]
    return uri_split(uri)[0]


def uri_grandparent(uri):
    return uri_parent(uri_parent(uri))


def uri_split(uri):
    return uri.rsplit('/', 1)


def uri_segment(uri, start=None, end=None):
    if start is None and end is None:
        return uri
    elif start is None:
        return '/'+'/'.join(uri.split('/')[:end])
    elif end is None:
        return '/'+'/'.join(uri.split('/')[start:])
    else:
        return '/'+'/'.join(uri.split('/')[start:end])


def uri_shape(uri):
    import re

    kwid_map = dict(zip(uri.split('/')[1::2], uri.split('/')[2::2]))
    shapes = {}

    for kw in kwid_map:
        seps = kwid_map[kw]

        for char in re.findall('[a-zA-Z0-9]', seps):
            seps = seps.replace(char, '')
            chunks = []
            p = '|'.join(seps)
            s = re.split(p, kwid_map[kw]) if p != '' else kwid_map[kw]
            for chunk in s:
                try:
                    float(chunk)
                    chunk = '*'
                except Exception:
                    pass

                chunks.append(chunk)

            shapes[kw] = '?'.join(chunks)

    return make_uri(shapes)


def make_uri(_dict):
    uri = ''

    kws = ['projects', 'subjects', 'experiments', 'assessors',
           'reconstructions', 'scans', 'resources', 'in_resources',
           'out_resources', 'files', 'in_files', 'out_files']

    for kw in kws:
        if kw in _dict.keys():
            uri += '/%s/%s' % (kw, _dict.get(kw))

    return uri


def check_entry(func):
    def inner(*args, **kwargs):
        args[0]._intf._get_entry_point()
        return func(*args, **kwargs)

    return inner


def extract_uri(uri):
    """
    Destructure the given REST uri into project,subject and experiment.

    Returns None if any one of project,subject or experiment is unspecified in
    the URI and a (project,subject,experiment) triple otherwise.
    """
    # elements in URLs are always separated by /, regardless of client
    split = uri.split('/')
    # a well qualified uri has a project subject, and experiment name
    # so when split the following items should be present:
    # ['', 'data', 'projects', 'project-name', 'subjects', 'subject-name',
    # 'experiments', 'experiment-name', 'scans']

    # Based on the above comment if there aren't 9 items in the split list the
    # uri isn't well qualified
    if (len(split) != 9):
        return None

    project = split[3]
    subject = split[5]
    experiment = split[7]

    return (project, subject, experiment)


def file_path(uri):
    """return the relative path of the file in the given URI

    for uri = '/.../files/a/b/c', return 'a/b/c'

    raises ValueError (through .index()) if '/files/' is not in the URI
    """
    return uri[7+uri.index('/files/'):]



import re

from lxml import etree
import six
if six.PY3:
    unicode = str

# parsing functions


def is_xnat_error(message):
    a = ['<!DOCTYPE', '<html>']
    try:
        return message.startswith(a[0]) or message.startswith(a[1])
    except TypeError:
        if isinstance(message, bytes):
            a = [bytes(e, 'utf-8') for e in a]
        return message.startswith(a[0]) or message.startswith(a[1])


def parse_error_message(message):
    try:
        if message.startswith('<html>'):
            message_tree = etree.XML(message)
            error_tag = message_tree.find('.//h3')
            if error_tag:
                error = error_tag.xpath("string()")
        elif message.startswith('<!DOCTYPE'):
            message_tree = etree.XML(message)
            error_tag = message_tree.find('.//title')
            if error_tag and 'Not Found' in error_tag.xpath("string()"):
                error = error_tag.xpath("string()")
            else:
                error_tag = message_tree.find('.//h1')
                if error_tag:
                    error = error_tag.xpath("string()")
        else:
            error = message

    except Exception:
        error = message
    finally:
        return error


def parse_put_error_message(message):
    error = parse_error_message(message)

    required_fields = []

    if error:
        for line in error.split('\n'):

            try:
                datatype_name = re.findall("\'.*?\'", line)[0].strip('\'')
                element_name = re.findall("\'.*?\'", line
                                          )[1].rsplit(':', 1)[1].strip('}\'')

                required_fields.append((datatype_name, element_name))
            except Exception:
                continue

    return required_fields


def catch_error(msg_or_exception, full_response=None):

    if isinstance(msg_or_exception, bytes):
        msg_or_exception = msg_or_exception.decode()

    # handle errors returned by the xnat server
    if isinstance(msg_or_exception, (str, unicode)):
        # parse the message
        msg = msg_or_exception
        error = parse_error_message(msg)

        # choose the exception
        if error == 'The request requires user authentication':
            raise OperationalError('Authentication failed')
        elif 'Not Found' in error:
            raise OperationalError('Connection failed')
        else:
            if full_response:
                raise DatabaseError(full_response)
            else:
                raise DatabaseError(error)

    # handle other errors, raised for instance by the http layer
    else:
        raise DatabaseError(str(msg_or_exception))


# Exceptions as defined in PEP-249, the module treats errors using thoses
# classes following as closely as possible the original definitions.


# http://python3porting.com/differences.html#standarderror
try:
    class Warning(StandardError):
        pass

    class Error(StandardError):
        pass
except NameError:
    class Warning(Exception):
        pass

    class Error(Exception):
        pass


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class DataError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class InternalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass



from pyxnat import select


def test_switch_to_singular():
    assert select.compute('/projects/nosetests') == ['/project/nosetests']


def test_switch_to_plural():
    assert select.compute('/project') == ['/projects/*']


def test_complete_stars_plural():
    assert select.compute('/projects/subjects/experiments') == \
           ['/projects/*/subjects/*/experiments/*']


def test_complete_stars_singular():
    assert select.compute('/project/subject/experiment') == \
           ['/projects/*/subjects/*/experiments/*']


def test_simple_root_expand():
    assert select.compute('//experiments') == \
           ['/projects/*/subjects/*/experiments/*']


def test_simple_level_expand():
    assert select.compute('/projects/IMAGEN//experiments') == \
           ['/project/IMAGEN/subjects/*/experiments/*']


def test_leaf_level_expand():
    assert set(select.compute('//files')) == \
        set(['/projects/*/subjects/*/experiments/*/resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/reconstructions/*/in_resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/scans/*/resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/assessors/*/out_resources/*/files/*',
             '/projects/*/subjects/*/resources/*/files/*',
             '/projects/*/resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/reconstructions/*/out_resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/assessors/*/in_resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/assessors/*/resources/*/files/*'])


def test_double_level_expand():
    assert set(select.compute('//experiments//files')) == \
        set(['/projects/*/subjects/*/experiments/*/resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/reconstructions/*/in_resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/scans/*/resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/assessors/*/out_resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/assessors/*/in_resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/reconstructions/*/out_resources/*/files/*',
             '/projects/*/subjects/*/experiments/*/assessors/*/resources/*/files/*'])


def test_compute_all():
    assert set(select.compute('/projects/nosetests//experiments/*Session*//files/myfile.txt')) == \
        set(['/project/nosetests/subjects/*/experiments/*Session*/resources/*/file/myfile.txt',
             '/project/nosetests/subjects/*/experiments/*Session*/reconstructions/*/in_resources/*/file/myfile.txt',
             '/project/nosetests/subjects/*/experiments/*Session*/scans/*/resources/*/file/myfile.txt',
             '/project/nosetests/subjects/*/experiments/*Session*/assessors/*/out_resources/*/file/myfile.txt',
             '/project/nosetests/subjects/*/experiments/*Session*/assessors/*/in_resources/*/file/myfile.txt',
             '/project/nosetests/subjects/*/experiments/*Session*/reconstructions/*/out_resources/*/file/myfile.txt',
             '/project/nosetests/subjects/*/experiments/*Session*/assessors/*/resources/*/file/myfile.txt'])



import re

from . import schema
from .search import Search
from .resources import CObject, Project, Projects, Experiment, Experiments
# imports used implicitly
from .uriutil import inv_translate_uri
# from .uriutil import uri_last
from .errors import ProgrammingError

DEBUG = False


def is_type_level(element):
    return element.strip('/') in schema.resources_types and \
           not is_expand_level(element)


def is_singular_type_level(element):
    return element.strip('/') in schema.resources_singular and \
           not is_expand_level(element)


def is_expand_level(element):
    return element.startswith('//') and \
        element.strip('/') in schema.resources_types


def is_id_level(element):
    return element is not None and \
        element.strip('/') not in schema.resources_types


def is_wildid_level(element):
    return element is not None and \
        element.strip('/') not in schema.resources_types and \
        ('?' in element or '*' in element)


def expand_level(element, fullpath):

    def find_paths(element, path=[]):
        resources_dict = schema.resources_tree

        element = element.strip('/')
        paths = []

        if path == []:
            path = [element]

        init_path = path[:]

        for key in resources_dict.keys():

            path = init_path[:]
            if element in resources_dict[key]:
                path.append(key)
                look_again = find_paths(key, path)

                if look_again != []:
                    paths.extend(look_again)
                else:
                    path.reverse()
                    paths.append('/' + '/'.join(path))

        return paths

    absolute_paths = find_paths(element)

    els = re.findall('/{1,2}.*?(?=/{1,2}|$)', fullpath)

    index = els.index(element)

    if index == 0:
        return absolute_paths
    else:
        for i in range(1, 4):
            if is_type_level(els[index - i]) or\
               is_expand_level(els[index - i]):
                parent_level = els[index - i]
                break

    if parent_level.strip('/') in schema.resources_singular:
        parent_level += 's'

    return [abspath.split(parent_level)[1]
            for abspath in absolute_paths
            if parent_level in abspath]


def mtransform(paths):
    tpaths = []

    for path in paths:
        els = re.findall('/{1,2}.*?(?=/{1,2}|$)', path)
        tels = []
        ignore_path = False

        for i, curr_el in enumerate(els):

            if i + 1 < len(els):
                next_el = els[i + 1]
            else:
                next_el = None

            if is_type_level(curr_el):

                if not is_id_level(next_el):
                    if not is_singular_type_level(curr_el):
                        tels.append(curr_el)
                        tels.append('/*')
                    else:
                        tels.append(curr_el + 's')
                        tels.append('/*')
                else:
                    if not is_singular_type_level(curr_el):
                        if not is_wildid_level(next_el):
                            tels.append(curr_el.rstrip('s'))
                        else:
                            tels.append(curr_el)
                    else:
                        if not is_wildid_level(next_el):
                            tels.append(curr_el)
                        else:
                            tels.append(curr_el + 's')

            elif is_expand_level(curr_el):

                exp_paths = [''.join(els[:i] + [rel_path] + els[i + 1:])
                             for rel_path in expand_level(curr_el, path)
                             ]

                tpaths.extend(mtransform(exp_paths))
                ignore_path = True
                break

            elif is_id_level(curr_el):
                tels.append(curr_el)

            else:
                raise ProgrammingError('in %s' % path)

        if not ignore_path:
            tpaths.append(''.join(tels))

    return tpaths


def group_paths(paths):
    groups = {}

    for path in paths:
        resources = [el
                     for el in re.findall('/{1,2}.*?(?=/{1,2}|$)', path)
                     if el.strip('/') in schema.resources_types
                     and el.strip('/') not in ['files', 'file']
                     ]

        if len(resources) == 1:
            groups.setdefault(resources[0], set()).add(path)
            continue

        for alt_path in paths:
            if alt_path.endswith(path):

                alt_rsc = \
                    [el for el in re.findall('/{1,2}.*?(?=/{1,2}|$)',
                                             alt_path
                                             )
                     if el.strip('/') in schema.resources_types
                        and el.strip('/') not in ['files', 'file']
                     ]

                if alt_rsc[-1].strip('/') in \
                        ['files', 'file', 'resources', 'resource'] + \
                        list(schema.rest_translation.keys()):

                    groups.setdefault(alt_rsc[-2] + alt_rsc[-1], set()
                                      ).add(alt_path)

                else:
                    groups.setdefault(alt_rsc[-1], set()).add(alt_path)

    return groups


def compute(path):
    if not re.match('/project(s)?|//.+', path):
        path = '/' + path

    path = inv_translate_uri(path)

    try:
        groups = group_paths(mtransform([path]))
    except Exception:
        raise ProgrammingError('in %s' % path)

    best = []

    for name in groups:
        lightest = (0, None)

        for path in groups[name]:
            score = len(path.split('/'))

            if lightest == (0, None) or lightest[0] > score:
                lightest = (score, path)

        best.append(lightest[1])

    return best


class Select(object):
    """ Data selection interface. Callable object that indicates the
        data to be returned to the user.

        Examples
        --------
            Select with a path:
                >>> interface.select('/projects/myproj/subjects').get()

            Select with a datatype:
                >>> columns = ['xnat:subjectData/PROJECT',
                               'xnat:subjectData/SUBJECT_ID'
                               ]
                >>> criteria = [('xnat:subjectData/SUBJECT_ID', 'LIKE', '*'),
                                'AND'
                                ]
                >>> interface.select('xnat:subjectData', columns
                            ).where(criteria)
    """
    def __init__(self, interface):
        """
            Parameters
            ----------
            interface: :class:`Interface`
                Main interface reference.
        """

        self._intf = interface

    def project(self, ID):
        """ Access a particular project.

            Parameters
            ----------
            ID: string
                ID of the project.
        """
        self._intf._get_entry_point()

        return globals()['Project'](
            '%s/projects/%s' % (self._intf._entry, ID), self._intf)

    def projects(self, id_filter='*'):
        """ Returns the list of all visible projects for the server.

            Parameters
            ----------
            id_filter: string
                Name pattern to filter the returned projects.
        """
        self._intf._get_entry_point()

        return globals()['Projects'](
            '%s/projects' % self._intf._entry, self._intf, id_filter)

    def experiment(self, ID):
        """ Access a particular experiment.

            Parameters
            ----------
            ID: string
                XNAT ID of experiment (NOT LABEL).
        """
        self._intf._get_entry_point()

        return globals()['Experiment'](
            '%s/experiments/%s' % (self._intf._entry, ID), self._intf)

    def experiments(self, id_filter='*'):
        """ Returns the list of all visible experiments for the server.

            Parameters
            ----------
            id_filter: string
                Name pattern to filter the returned experiments
                (MUST BE XNAT ID's not labels).
        """
        self._intf._get_entry_point()

        return globals()['Experiments'](
            '%s/experiments' % self._intf._entry, self._intf, id_filter)

    def tag(self, name):
        self._intf._get_entry_point()

        return self._intf.manage.tags.get(name).references()

    def tags(self):
        self._intf._get_entry_point()

        return self._intf.manage.tags()

    def __repr__(self):
        return '<Root Object>'

    def __call__(self, datatype_or_path, columns=[]):
        """ Select clause to specify what type of data is to be returned.

            Parameters
            ----------
            datatype_or_path: string
                Can either be a resource path or a datatype:
                    - when a path, REST resources are returned, the
                      `columns` argument is useless.
                    - when a datatype, a search Object is returned,
                      the `columns` argument has to be specified.
            columns: list
                List of fieldtypes e.g. xnat:subjectData/SUBJECT_ID
                Datatype and columns are used to specify the search table
                that has to be returned. Use the method `where` on the
                `Search` object to trigger a search on the database.
        """
        self._intf._get_entry_point()

        if datatype_or_path.startswith('/tag'):
            if len(datatype_or_path.split('/')) == 3:
                return self.tag(datatype_or_path.split('/')[-1])
            else:
                return self.tags()

        if datatype_or_path in ['/', '//', self._intf._entry]:
            return self

        if datatype_or_path.startswith(self._intf._entry):
            datatype_or_path = datatype_or_path.split(
                self._intf._entry, 1)[1]

        if datatype_or_path.startswith('/'):
            return_list = []

            try:
                for path in compute(datatype_or_path):
                    if DEBUG:
                        print('path: %s' % path)

                    pairs = zip(path.split('/')[1::2], path.split('/')[2::2])

                    # # in case a level id has a / - allowed for files only
                    # if len(path.split('/')[1:]) % 2 == 1 \
                    #         and uri_last(path) not in schema.resources_types:

                    #     pairs[-1] = (pairs[-1][0], uri_last(path))

                    obj = self
                    for resource, identifier in pairs:

                        if isinstance(obj, list):
                            obj = [getattr(sobj, resource)(identifier)
                                   for sobj in obj]
                        else:
                            obj = getattr(obj, resource)(identifier)

                    return_list.append(obj)

                if len(return_list) == 1:
                    return return_list[0]
                else:
                    return CObject(return_list, self._intf)

            except Exception as e:
                if DEBUG:
                    print(e)
                raise ProgrammingError('in %s' % datatype_or_path)

        else:
            if columns == []:
                columns = self._intf.inspect.datatypes(datatype_or_path)

            return Search(datatype_or_path, columns, self._intf)



from pyxnat import Interface
from pyxnat.core import downloadutils as du
import os.path as op
from . import skip_if_no_network

fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
central = Interface(config=fp)


@skip_if_no_network
def test_download():
    s = central.select.project('CENTRAL_OASIS_LONG').subject('OAS2_0001')
    e = s.experiment('CENTRAL_E00090')
    du.download(dest_dir='/tmp',
                instance=e.scans(),
                extract=True,
                removeZip=True)



from pyxnat.core.help import GraphData, PaintGraph, SchemasInspector
from pyxnat import Interface
from . import skip_if_no_network
import os.path as op

fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
central = Interface(config=fp)


@skip_if_no_network
def test_graphdata():
    g = GraphData(central)
    g.architecture()
    g.datatypes()
    g.rest_resource('projects')


@skip_if_no_network
def test_schemasinspector():
    si = SchemasInspector(central)
    si()


def test_paintgraph():
    PaintGraph(central)



import os.path as op
from pyxnat import Interface

fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
central = Interface(config=fp)
p = central.select.project('nosetests5')
s = p.subject('rs')
e1 = s.experiment('CENTRAL02_E01603')

p = central.select.project('surfmask_smpl2')
s = p.subject('CENTRAL05_S01120')
e2 = s.experiment('CENTRAL05_E02681')


def test_ashs_volumes():

    r = e1.resource('ASHS')
    hv = r.volumes()
    v = hv.query('region=="CA1" & side=="left"')['volume'].tolist()[0]
    assert(v == 1287.675)


def test_freesurfer6_aparc():
    r = e1.resource('FREESURFER6')
    hv = r.aparc()
    q = 'region=="supramarginal" & side=="left" & measurement=="CurvInd"'
    v = hv.query(q)['value'].tolist()[0]
    assert(v == '10.8')


def test_freesurfer6_aseg():
    r = e1.resource('FREESURFER6')
    hv = r.aseg()
    v = hv.query('region=="BrainSegVol"')['value'].tolist()[0]
    assert(v == 2463095.0)


def test_scandate():
    s = e2.scans().first()
    r = s.resource('DICOM')
    assert(r.scandate() == '2008-05-06')



from pyxnat import uriutil


def test_translate_uri():
    assert uriutil.translate_uri('/assessors/out_resources/files') == \
           '/assessors/out/resources/files'
    assert uriutil.translate_uri('/assessors/out_resource/files') == \
           '/assessors/out/resource/files'
    assert uriutil.translate_uri('/assessors/in_resources/files') == \
           '/assessors/in/resources/files'
    assert uriutil.translate_uri('/assessors/in_resource/files') == \
           '/assessors/in/resource/files'


def test_inv_translate_uri():
    assert uriutil.inv_translate_uri('/assessors/out/resources/files') == \
           '/assessors/out_resources/files'
    assert uriutil.inv_translate_uri('/assessors/out/resource/files') == \
           '/assessors/out_resource/files'
    assert uriutil.inv_translate_uri('/assessors/in/resources/files') == \
           '/assessors/in_resources/files'
    assert uriutil.inv_translate_uri('/assessors/in/resource/files') == \
           '/assessors/in_resource/files'


def test_join_uri():
    uri = uriutil.join_uri('/projects', 'project_id', 'subjects', 'subject_id')
    assert uri == '/projects/project_id/subjects/subject_id'


def test_uri_last():
    assert uriutil.uri_last('/projects/1/subjects/2') == '2'


def test_uri_nextlast():
    assert uriutil.uri_nextlast('/projects/1/subjects/2') == 'subjects'


def test_uri_parent():
    uri = uriutil.uri_parent('/projects/1/subjects/2')
    assert uri == '/projects/1/subjects'


def test_uri_grandparent():
    uri = uriutil.uri_grandparent('/projects/1/subjects/2')
    assert uri == '/projects/1'


def test_uri_split():
    uri = uriutil.uri_split('/projects/1/subjects/2')
    assert uri == ['/projects/1/subjects', '2']



#!/usr/bin/env python
'''
The MIT License (MIT)

Copyright (c) 2014 VUIIS

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

'''
Adapted from xnatmirror (https://github.com/VUIIS/dax/blob/master/bin/Xnat_tools/Xnatmirror)
'''

def cmp(a, b):
    return (a > b) - (a < b)

from builtins import zip
from builtins import str

import os
import os.path as op
import sys

from xml.etree import cElementTree as ET
import pyxnat

PROJ_ATTRS = [
    'xnat:projectData/name',
    'xnat:projectData/description',
    'xnat:projectData/keywords',
]

SUBJ_ATTRS = [
    'xnat:subjectData/group',
    'xnat:subjectData/src',
    'xnat:subjectData/investigator/firstname',
    'xnat:subjectData/investigator/lastname',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/dob',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/yob',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/age',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/gender',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/handedness',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/ses',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/education',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/\
educationDesc',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/race',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/ethnicity',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/weight',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/height',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/\
gestational_age',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/\
post_menstrual_age',
    'xnat:subjectData/demographics[@xsi:type=xnat:demographicData]/\
birth_weight'
]

MR_EXP_ATTRS = [
    'xnat:experimentData/date',
    'xnat:experimentData/visit_id',
    'xnat:experimentData/time',
    'xnat:experimentData/note',
    'xnat:experimentData/investigator/firstname',
    'xnat:experimentData/investigator/lastname',
    'xnat:imageSessionData/scanner/manufacturer',
    'xnat:imageSessionData/scanner/model',
    'xnat:imageSessionData/operator',
    'xnat:imageSessionData/dcmAccessionNumber',
    'xnat:imageSessionData/dcmPatientId',
    'xnat:imageSessionData/dcmPatientName',
    'xnat:imageSessionData/session_type',
    'xnat:imageSessionData/modality',
    'xnat:imageSessionData/UID',
    'xnat:mrSessionData/coil',
    'xnat:mrSessionData/fieldStrength',
    'xnat:mrSessionData/marker',
    'xnat:mrSessionData/stabilization'
]

OTHER_DICOM_SCAN_ATTRS = [
    'xnat:imageScanData/type',
    'xnat:imageScanData/UID',
    'xnat:imageScanData/note',
    'xnat:imageScanData/quality',
    'xnat:imageScanData/condition',
    'xnat:imageScanData/series_description',
    'xnat:imageScanData/documentation',
    'xnat:imageScanData/frames',
    'xnat:imageScanData/startTime',
    'xnat:imageScanData/scanner/manufacturer',
    'xnat:imageScanData/scanner/model'
]

MR_SCAN_ATTRS = [
    'xnat:imageScanData/type',
    'xnat:imageScanData/UID',
    'xnat:imageScanData/note',
    'xnat:imageScanData/quality',
    'xnat:imageScanData/condition',
    'xnat:imageScanData/series_description',
    'xnat:imageScanData/documentation',
    'xnat:imageScanData/frames',
    'xnat:imageScanData/startTime',
    'xnat:imageScanData/scanner/manufacturer',
    'xnat:imageScanData/scanner/model',
    'xnat:mrScanData/parameters/flip',
    'xnat:mrScanData/parameters/orientation',
    'xnat:mrScanData/parameters/tr',
    'xnat:mrScanData/parameters/ti',
    'xnat:mrScanData/parameters/te',
    'xnat:mrScanData/parameters/sequence',
    'xnat:mrScanData/parameters/imageType',
    'xnat:mrScanData/parameters/scanSequence',
    'xnat:mrScanData/parameters/seqVariant',
    'xnat:mrScanData/parameters/scanOptions',
    'xnat:mrScanData/parameters/acqType',
    'xnat:mrScanData/parameters/pixelBandwidth',
    'xnat:mrScanData/parameters/voxelRes/x',
    'xnat:mrScanData/parameters/voxelRes/y',
    'xnat:mrScanData/parameters/voxelRes/z',
    'xnat:mrScanData/parameters/fov/x',
    'xnat:mrScanData/parameters/fov/y',
    'xnat:mrScanData/parameters/matrix/x',
    'xnat:mrScanData/parameters/matrix/y',
    'xnat:mrScanData/parameters/partitions',
    'xnat:mrScanData/fieldStrength',
    'xnat:mrScanData/marker',
    'xnat:mrScanData/stabilization',
    'xnat:mrScanData/coil'
]

SC_SCAN_ATTRS = [
    'xnat:imageScanData/type',
    'xnat:imageScanData/UID',
    'xnat:imageScanData/note',
    'xnat:imageScanData/quality',
    'xnat:imageScanData/condition',
    'xnat:imageScanData/series_description',
    'xnat:imageScanData/documentation',
    'xnat:imageScanData/frames',
    'xnat:imageScanData/scanner/manufacturer',
    'xnat:imageScanData/scanner/model'
]

PET_EXP_ATTRS = [
    'xnat:experimentData/date',
    'xnat:experimentData/visit_id',
    'xnat:experimentData/time',
    'xnat:experimentData/note',
    'xnat:experimentData/investigator/firstname',
    'xnat:experimentData/investigator/lastname',
    'xnat:imageSessionData/scanner/manufacturer',
    'xnat:imageSessionData/scanner/model',
    'xnat:imageSessionData/operator',
    'xnat:imageSessionData/dcmAccessionNumber',
    'xnat:imageSessionData/dcmPatientId',
    'xnat:imageSessionData/dcmPatientName',
    'xnat:imageSessionData/session_type',
    'xnat:imageSessionData/modality',
    'xnat:imageSessionData/UID',
    'xnat:petSessionData/studyType',
    'xnat:petSessionData/patientID',
    'xnat:petSessionData/patientName',
    'xnat:petSessionData/stabilization',
    'xnat:petSessionData/start_time/scan',
    'xnat:petSessionData/start_time/injection',
    'xnat:petSessionData/tracer/name',
    'xnat:petSessionData/tracer/startTime',
    'xnat:petSessionData/tracer/dose',
    'xnat:petSessionData/tracer/specificActivity',
    'xnat:petSessionData/tracer/totalMass',
    'xnat:petSessionData/tracer/intermediate',
    'xnat:petSessionData/tracer/isotope',
    'xnat:petSessionData/tracer/isotope/half-life',
    'xnat:petSessionData/tracer/transmissions',
    'xnat:petSessionData/tracer/transmissions/startTime'
]

CT_EXP_ATTRS = [
    'xnat:experimentData/date',
    'xnat:experimentData/visit_id',
    'xnat:experimentData/time',
    'xnat:experimentData/note',
    'xnat:experimentData/investigator/firstname',
    'xnat:experimentData/investigator/lastname',
    'xnat:imageSessionData/scanner/manufacturer',
    'xnat:imageSessionData/scanner/model',
    'xnat:imageSessionData/operator',
    'xnat:imageSessionData/dcmAccessionNumber',
    'xnat:imageSessionData/dcmPatientId',
    'xnat:imageSessionData/dcmPatientName',
    'xnat:imageSessionData/session_type',
    'xnat:imageSessionData/modality',
    'xnat:imageSessionData/UID'
]

PET_SCAN_ATTRS = [
    'xnat:imageScanData/type',
    'xnat:imageScanData/UID',
    'xnat:imageScanData/note',
    'xnat:imageScanData/quality',
    'xnat:imageScanData/condition',
    'xnat:imageScanData/series_description',
    'xnat:imageScanData/documentation',
    'xnat:imageScanData/frames',
    'xnat:imageScanData/scanner/manufacturer',
    'xnat:imageScanData/scanner/model',
    'xnat:imageScanData/startTime',
    'xnat:petScanData/parameters/orientation',
    'xnat:petScanData/parameters/originalFileName',
    'xnat:petScanData/parameters/systemType',
    'xnat:petScanData/parameters/fileType',
    'xnat:petScanData/parameters/transaxialFOV',
    'xnat:petScanData/parameters/acqType',
    'xnat:petScanData/parameters/facility',
    'xnat:petScanData/parameters/numPlanes',
    'xnat:petScanData/parameters/frames/numFrames',
    'xnat:petScanData/parameters/numGates',
    'xnat:petScanData/parameters/planeSeparation',
    'xnat:petScanData/parameters/binSize',
    'xnat:petScanData/parameters/dataType'
]

CT_SCAN_ATTRS = [
    'xnat:imageScanData/type',
    'xnat:imageScanData/UID',
    'xnat:imageScanData/note',
    'xnat:imageScanData/quality',
    'xnat:imageScanData/condition',
    'xnat:imageScanData/series_description',
    'xnat:imageScanData/documentation',
    'xnat:imageScanData/frames',
    'xnat:imageScanData/scanner/manufacturer',
    'xnat:imageScanData/scanner/model'
]

PROC_ATTRS = [
    'proc:genProcData/validation/status',
    'proc:genProcData/procstatus',
    'proc:genProcData/proctype',
    'proc:genProcData/procversion',
    'proc:genProcData/walltimeused',
    'proc:genProcData/memused'
]


def copy_attrs(src_obj, dest_obj, attr_list):
    """ Copies list of attributes form source to destination"""
    src_attrs = src_obj.attrs.mget(attr_list)
    src_list = dict(list(zip(attr_list, src_attrs)))

    # NOTE: For some reason need to set te again b/c a bug somewhere sets te
    # to sequence name
    te_key = 'xnat:mrScanData/parameters/te'
    if te_key in src_list:
        src_list[te_key] = src_obj.attrs.get(te_key)

    dest_obj.attrs.mset(src_list)
    return 0


def copy_attributes(src_obj, dest_obj):
    '''Copy attributes from src to dest'''
    src_type = src_obj.datatype()
    types = {'xnat:projectData': PROJ_ATTRS,
             'xnat:subjectData': SUBJ_ATTRS,
             'xnat:mrSessionData': MR_EXP_ATTRS,
             'xnat:petSessionData': PET_EXP_ATTRS,
             'xnat:ctSessionData': CT_EXP_ATTRS,
             'xnat:mrScanData': MR_SCAN_ATTRS,
             'xnat:petScanData': PET_SCAN_ATTRS,
             'xnat:ctScanData': CT_SCAN_ATTRS,
             'xnat:scScanData': SC_SCAN_ATTRS,
             'proc:genProcData': PROC_ATTRS,
             'xnat:otherDicomScanData': OTHER_DICOM_SCAN_ATTRS}

    try:
        copy_attrs(src_obj, dest_obj, types[src_type])
    except KeyError:
        print('ERROR:cannot copy attributes, unsupported datatype:' + src_type)


def subj_compare(item1, item2):
    '''Compare sort of items'''
    return cmp(item1.label(), item2.label())


def copy_file(src_f, dest_r, cache_d):
    '''
    Copy file from XNAT file source to XNAT resource destination,
    using local cache in between'''
    f_label = src_f.label()
    loc_f = cache_d + '/' + f_label

    # Make subdirectories
    loc_d = op.dirname(loc_f)
    if not op.exists(loc_d):
        os.makedirs(loc_d)

    try:
        # Download file
        if op.exists(loc_f) is False:
            src_f.get(loc_f)

        # Get File Attributes
        f_in_attrs = src_f.attributes()
        f_content = f_in_attrs.get('file_content')
        f_format = f_in_attrs.get('file_format')
        f_tags = f_in_attrs.get('file_tags')

        # Upload File
        if f_format and f_content and not f_tags:         # format & content
            dest_r.file(f_label).put(loc_f, f_format, f_content)
        elif f_format and not f_content and not f_tags:   # format only
            dest_r.file(f_label).put(loc_f, f_format)
        elif f_format and f_content and f_tags:   # format, content, & tags
            dest_r.file(f_label).put(loc_f, f_format, f_content)
        else:                                             # none
            dest_r.file(f_label).put(loc_f)

        # Delete local copy
        os.remove(loc_f)
    except Exception:
        print("ERROR:failed to copy file:%s, error=%s"
              % (f_label, sys.exc_info()[0]))


def copy_res_zip(src_r, dest_r, cache_d):
    '''
    Copy a resource from XNAT source to XNAT destination using local cache
    in between
    '''
    try:
        # Download zip of resource
        print('INFO:Downloading resource as zip...')
        cache_z = src_r.get(cache_d, extract=False)

        # Upload zip of resource
        print('INFO:Uploading resource as zip...')
        dest_r.put_zip(cache_z, extract=True)

        # Delete cached zip
        os.remove(cache_z)

    except IndexError:
        print('ERROR:failed to copy:%s:%s' % (cache_z, sys.exc_info()[0]))
        raise


def is_empty_resource(_res):
    '''Check if resource contains any files'''
    f_count = 0
    for f_in in _res.files().fetchall('obj'):
        f_count += 1
        break

    return f_count == 0

# copy_project and copy_subject are untested

# def copy_project(src_proj, dst_proj, proj_cache_dir):
#     '''Copy XNAT project from source to destination'''
#
#     if not dst_proj.exists():
#         try:
#             dst_proj.create()
#         except Exception as e:
#             msg = 'ERROR: can not create project on destination '\
#                 'xnat. Check your user rights. %s' % e
#             raise Exception(msg)
#         copy_attributes(src_proj, dst_proj)
#     print('INFO:loading and sorting subject list...')
#
#     proj_label = src_proj.label()
#     subj_list = src_xnat.get_subjects(proj_label)
#     subj_list = [x for x in subj_list if x['label'] not in SUBJECTS_MIRRORED]
#     subj_i = 0
#     for subj in subj_list:
#         subj_i += 1
#         subject_label = subj['label']
#
#         print("INFO:Processing subject %s (%d)..." % (subject_label, subj_i))
#         src_subj = src_proj.subject(subject_label)
#         dst_subj = dst_proj.subject(subject_label)
#         subj_cache_dir = op.join(proj_cache_dir, subject_label)
#         copy_subject(src_subj, dst_subj, subj_cache_dir)


# def copy_subject(src_subj, dst_subj, subj_cache_dir):
#     '''Copy subject from XNAT src to XNAT dst'''
#
#     if not dst_subj.exists():
#         print('INFO:uploading subject attributes as xml')
#
#         # Create dirs
#         if not op.exists(subj_cache_dir):
#             os.makedirs(subj_cache_dir)
#
#         # Write xml to file
#         subj_xml = src_subj.get()
#         xml_path = op.join(subj_cache_dir, 'subj.xml')
#         write_xml(subj_xml, xml_path)
#         dst_subj.create(xml=xml_path, allowDataDeletion=False)
#
#     # Process each experiment of subject
#     for src_sess in src_subj.experiments().fetchall('obj'):
#         sess_label = src_sess.label()
#         sess_type = src_sess.datatype()
#         if sess_type != 'xnat:mrSessionData' and \
#            sess_type != 'xnat:petSessionData' and \
#            sess_type != 'xnat:ctSessionData':
#             print('WARN:Skipping, session is not MR, CT, or PET Session')
#             continue
#
#         print("INFO:Processing session:%s..." % (sess_label))
#
#         dst_sess = dst_subj.experiment(sess_label)
#         sess_cache_dir = op.join(subj_cache_dir, sess_label)
#         copy_session(src_sess, dst_sess, sess_cache_dir)


def copy_session(src_sess, dst_sess, sess_cache_dir):
    '''Copy XNAT session from source to destination'''

    print(dst_sess.exists())
    print(dst_sess._uri)
    print('INFO:uploading session attributes as xml')
    # Write xml to file
    if not op.exists(sess_cache_dir):
        os.makedirs(sess_cache_dir)
    sess_xml = src_sess.get()
    xml_path = op.join(sess_cache_dir, 'sess.xml')
    write_xml(sess_xml, xml_path)
    print(xml_path)

    sess_type = src_sess.datatype()
    dst_sess.create(experiments=sess_type)
    copy_attributes(src_sess, dst_sess)

    # Process each scan of session
    for src_scan in src_sess.scans().fetchall('obj'):
        scan_label = src_scan.label()

        print('INFO:Processing scan:%s...' % scan_label)
        dst_scan = dst_sess.scan(scan_label)
        scan_cache_dir = op.join(sess_cache_dir, scan_label)
        copy_scan(src_scan, dst_scan, scan_cache_dir)

    # Process each assessor of session
    for src_assr in src_sess.assessors():
        assr_label = src_assr.label()
        print('INFO:Processing assessor:%s:...' % assr_label)
        dst_assr = dst_sess.assessor(assr_label)
        assr_cache_dir = op.join(sess_cache_dir, assr_label)
        copy_assr(src_assr, dst_assr, assr_cache_dir)

    for src_res in src_sess.resources().fetchall('obj'):
        res_label = src_res.label()

        print('INFO:Processing resource:%s...' % (res_label))

        dst_res = dst_sess.resource(res_label)

        res_cache_dir = op.join(scan_cache_dir, res_label)

        copy_res(src_res, dst_res, res_cache_dir, use_zip=True)


def copy_scan(src_scan, dst_scan, scan_cache_dir):
    '''Copy scan from source XNAT to destination XNAT'''

    scan_type = src_scan.datatype()
    if scan_type == '':
             scan_type = 'xnat:otherDicomScanData'
    dst_scan.create(scans=scan_type)
    copy_attributes(src_scan, dst_scan)

    # Process each resource of scan
    for src_res in src_scan.resources().fetchall('obj'):
        res_label = src_res.label()

        print('INFO:Processing resource:%s...' % (res_label))

        dst_res = dst_scan.resource(res_label)

        res_cache_dir = op.join(scan_cache_dir, res_label)
        if res_label == 'SNAPSHOTS':
            copy_res(src_res, dst_res, res_cache_dir)
        else:
            copy_res(src_res, dst_res, res_cache_dir, use_zip=True)


def copy_res(src_res, dst_res, res_cache_dir, use_zip=False):
    '''Copy resource from source XNAT to destination XNAT'''
    # Create cache dir
    if not op.exists(res_cache_dir):
        os.makedirs(res_cache_dir)

    # Prepare resource and check for empty
    is_empty = False
    print(dst_res._uri)
    if not dst_res.exists():
        dst_res.create()
        is_empty = True
    elif is_empty_resource(dst_res):
        is_empty = True

    # Check for empty source
    if is_empty_resource(src_res):
        print('WARN:empty resource, nothing to copy')
        return

    if is_empty:
        if use_zip:
            # Try to copy as zip
            try:
                print('INFO:Copying resource as zip: %s...' % src_res.label())
                copy_res_zip(src_res, dst_res, res_cache_dir)
                return
            except Exception:
                try:
                    print('INFO: second attempt to copy resource as zip: %s...'
                          % src_res.label())
                    copy_res_zip(src_res, dst_res, res_cache_dir)
                    return
                except Exception:
                    msg = 'ERROR:failed twice to copy resource as zip, will'\
                        'copy individual files'
                    print(msg)

        copy_count = 0
        for f in src_res.files():
            print('INFO:Copying file: %s...' % f.label())
            copy_count += 1
            copy_file(f, dst_res, res_cache_dir)
        print('INFO:Finished copying resource, %d files copied' % copy_count)


# def copy_assr(src_assr, dst_assr, assr_cache_dir):
#     '''Copy assessor from source XNAT to destination XNAT'''
#
#     # Check type
#     assr_type = src_assr.datatype()
#     if assr_type != 'proc:genProcData' and assr_type != 'fs:fsData':
#         print('WARN:skipping unsupported assessor type: {}'.format(assr_type))
#         return
#
#     if not dst_assr.exists():
#         print('INFO:uploading assessor attributes as xml')
#         # Write xml to file
#         if not op.exists(assr_cache_dir):
#             os.makedirs(assr_cache_dir)
#         assr_xml = src_assr.get()
#         xml_path = op.join(assr_cache_dir, 'assr.xml')
#         write_xml(assr_xml, xml_path)
#         dst_assr.create(xml=xml_path, allowDataDeletion=False)
#
#     # Process each resource of assr
#     for src_res in src_assr.out_resources():
#         res_label = src_res.label()
#         print('INFO:Processing resource:%s...' % res_label)
#         dst_res = dst_assr.out_resource(res_label)
#         res_cache_dir = op.join(assr_cache_dir, res_label)
#
#         if res_label == 'SNAPSHOTS':
#             copy_res(src_res, dst_res, res_cache_dir)
#         else:
#             copy_res(src_res, dst_res, res_cache_dir, use_zip=True)


def write_xml(xml_str, file_path, clean_tags=True):
    """Writing XML."""
    root = ET.fromstring(xml_str)

    # We only want the tags and attributes relevant to root, no children
    if clean_tags:
        # Remove ID
        if 'ID' in root.attrib:
            del root.attrib['ID']

        # Remove sharing tags
        tag = '{http://nrg.wustl.edu/xnat}sharing'
        for child in root.findall(tag):
            root.remove(child)

        # Remove out
        for child in root.findall('{http://nrg.wustl.edu/xnat}out'):
            root.remove(child)
            break

        # Remove session ID
        tag = '{http://nrg.wustl.edu/xnat}imageSession_ID'
        for child in root.findall(tag):
            root.remove(child)

        # Remove subject ID
        for child in root.findall('{http://nrg.wustl.edu/xnat}subject_ID'):
            root.remove(child)

        # Remove _session ID
        tag = '{http://nrg.wustl.edu/xnat}image_session_ID'
        for child in root.findall(tag):
            root.remove(child)

        # Remove scans
        for child in root.findall('{http://nrg.wustl.edu/xnat}scans'):
            root.remove(child)
            break

        # Remove assessors
        for child in root.findall('{http://nrg.wustl.edu/xnat}assessors'):
            root.remove(child)
            break

        # Remove resources
        for child in root.findall('{http://nrg.wustl.edu/xnat}resources'):
            root.remove(child)
            break

        # Remove experiments
        for child in root.findall('{http://nrg.wustl.edu/xnat}experiments'):
            root.remove(child)
            break

    try:
        # Write to file
        ET.register_namespace('xnat', 'http://nrg.wustl.edu/xnat')
        ET.register_namespace('proc', 'http://nrg.wustl.edu/proc')
        ET.register_namespace('prov', 'http://www.nbirn.net/prov')
        ET.register_namespace('fs', 'http://nrg.wustl.edu/fs')
        ET.ElementTree(root).write(file_path)
    except IOError as error:
        print('ERROR:writing xml file: {}: {}'.format(file_path, str(error)))


def create_parser():
    import argparse
    """Parse commandline arguments."""
    parser = argparse.ArgumentParser(description='Downloads a given experiment/'\
        'session from an XNAT instance and uploads it to an independent one. '\
        'Only DICOM resources will be imported. \n\n'\
        'More information at: https://wiki.xnat.org/docs16/4-developer-documen'\
        'tation/xnat-rest-api-directory/data-services-import',
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--h1', '--source_config', dest='source_config',
        help='Source XNAT configuration file', required=True)
    parser.add_argument('--h2', '--dest_config', dest='dest_config',
        help='Destination XNAT configuration file', required=True)
    parser.add_argument('-e', '--experiment_id',
        help='Which resource to download? (Entity name/identifier)', required=True)
    parser.add_argument('-p', '--project_id', dest='project_id',
        help='Which project to store the resource in', required=True)

    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
        default=False, help='Display verbosal information (optional)',
        required=False)

    return parser



def main(args):
    x1 = pyxnat.Interface(config=args.source_config)
    x2 = pyxnat.Interface(config=args.dest_config)

    columns = ['subject_label', 'label']
    e1 = x1.array.experiments(experiment_id=args.experiment_id,
        columns=columns).data[0]
    p = x2.select.project(args.project_id)
    s = p.subject(e1['subject_label'])
    if not s.exists():
       s.create()
    e = s.experiment(e1['label'])
    e.create()

    src_sess = x1.select.project(e1['project']).subject(e1['subject_ID']).experiment(e1['ID'])
    dst_sess = e

    copy_session(src_sess, dst_sess, '/tmp')


if __name__ == '__main__':
    parser = create_parser()
    args = parser.parse_args()
    main(args)



import re
from lxml import etree

from .jsonutil import JsonTable


def get_subject_id(location):
    f = open(location, 'rb')
    content = f.read()
    f.close()

    subject = re.findall('<xnat:Subject\sID="(.*?)"\s', content)

    if subject != []:
        return subject[0]


class XpathStore(object):

    """ This utility class is used to query XML files describing XNAT
        subjects but which are stored in the cache directory. Xpath is
        used as the query lanaguage but a few helpers are provided
        to make simple queries.
    """

    def __init__(self, interface):
        self._intf = interface
        self._nsmap = {}
        self._tree = etree.Element('Store')

    def __call__(self, xpath):
        if self._tree is None:
            self._load()

        return self._tree.xpath(xpath, namespaces=self._nsmap)

    def _load(self, content):
        if not content:
            return
        new_subject = etree.fromstring(content)
        self._nsmap = new_subject.nsmap
        subject_id = new_subject.xpath('@ID')[0]
        old_subject = self.subject(subject_id)
        if old_subject:
            self._tree.replace(old_subject, new_subject)
        else:
            self._tree.append(new_subject)

    def _last_modified(self):
        entry_point = self._intf._get_entry_point()
        uri = '%s/subjects?columns=last_modified' % entry_point

        return dict(JsonTable(self._intf._get_json(uri),
                              order_by=['ID', 'last_modified']
                              ).select(['ID', 'last_modified']
                                       ).items())

    def update(self):
        """ Updates the xml files in the cachedir.
        """

        self.checkout(subjects=self.subjects())

    def checkout(self, project=None, subjects=None):
        """ Downloads all the subject XMLs for a project or a list
            of subjects.

            Parameters
            ----------
            project: str
                The id of the project
            subjects: list
                List of subjects that have to be downloaded.
        """
        if project is not None and subjects is None:
            for s in self._intf.select('/project/%s/subjects' % project):
                self._load(s.get())
        elif subjects is not None:
            for sid in subjects:
                self._load(self._intf._exec('%s/subjects/%s?format=xml' % (
                          self._intf._get_entry_point(), sid))
                          )

    def subject(self, subject_id):
        tmp = self.__call__('//xnat:Subject[@ID="%s"]' % (subject_id))

        if len(tmp) > 0:
            return tmp[1]
        return None

    def subjects(self):
        """ Returns all the subject ids.
        """
        return self.__call__('//xnat:Subject/attribute::ID')

    def keys(self):
        """ Returns attribute keys at the subject level.
        """
        return self.element_keys('xnat:Subject')

    def values(self, key):
        """ Returns the values for all subjects for a specific attribute.
        """
        return self.element_values('xnat:Subject', key)

    def attrs(self):
        """ Returns the attributes of all subjects.
        """
        return self.element_attrs('xnat:Subject')

    def elements(self):
        """ Returns the element names for all subjects elements.

            .. note:: in xpath terms, this function is returning all
                the names of the subjects descendant nodes.
        """
        elements = set()

        for element in self.__call__('//xnat:Subject/descendant::*'):
            for ns in element.nsmap:
                if element.nsmap[ns] in element.tag:
                    n = element.tag.replace('{%s}' % element.nsmap[ns],
                                            '%s:' % ns)
                    elements.add(n)

        return list(elements)

    def element_attrs(self, name):
        """ Returns the attributes of this specific element.
        """
        attrs = []

        for element in self.__call__('//%s' % name):
            attrs.append(element.attrib)

        return attrs

    def element_keys(self, name):
        """ Returns the attribute keys of this specific element.
        """
        keys = set()

        for element in self.__call__('//%s' % name):
            for element_key in element.keys():
                keys.add(element_key)

        return list(keys)

    def element_values(self, name, key):
        """ Returns the attribute values of this specific element.
        """
        values = set()

        for subject in self.__call__('//%s' % name):
            values.add(subject.get(key))

        return list(values)

    def element_text(self, name):
        """ Returns the text values of this specific element.
        """
        return list(set(self.__call__('//%s/child::text()' % name)))



import os.path as op
from pyxnat import Interface


fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
central = Interface(config=fp)


def test_pipelines_get():
    from pyxnat.core import pipelines
    p = pipelines.Pipelines('nosetests5', central)
    print(p.get())



XNAT_RESOURCE_NAME = 'CAT12_SEGMENT'


def volumes(self):
    import tempfile
    import nibabel as nib
    import numpy as np

    vols = []
    _, fp = tempfile.mkstemp(suffix='.nii.gz')

    names = ['mri/p1', 'mri/p2', 'mri/p3']
    for kls in names:
        f = [each for each in self.files() if each.id().startswith(kls)][0]
        f.get(fp)
        d = nib.load(fp)
        size = np.prod(d.header['pixdim'].tolist()[:4])
        v = np.sum(d.dataobj) * size
        vols.append((kls, v))

    return dict(vols)


def download_rc(self, path):
    '''Downloads a local copy of DARTEL imports. (Keep in mind that in CAT12
    images are named rp*, not rc*.'''
    import os.path as op
    rc_files = ['mri/rp1', 'mri/rp2']
    for each in rc_files:
        res = self._intf.array.experiments(experiment_id=self.parent().id(),
                                           columns=['subject_label']).data[0]
        subject_label = res['subject_label']
        id = res['ID']
        fp = '%s_%s_%s.nii.gz' % (each[-3:], subject_label, id)
        list(self.files('%s*' % each))[0].get(op.join(path, fp))



try:
    import networkx as nx
    from networkx.drawing.nx_agraph import graphviz_layout
    import matplotlib.pyplot as plt
    _DRAW_GRAPHS = True
except Exception:
    _DRAW_GRAPHS = False

from . import schema
from .jsonutil import get_column
from .search import Search


class Inspector(object):
    """ Database introspection interface.
    """

    def __init__(self, interface):
        """
            Parameters
            ----------
            interface:
                :class:`Interface` Object
        """
        self._intf = interface
        self._get_json = interface._get_json
        self._auto = True
        self._tick = 30

        self.schemas = SchemasInspector(interface)

    def __call__(self):
        for name in ['experiment', 'assessor', 'scan', 'reconstruction']:
            self._resource_struct(name)

    def set_autolearn(self, auto=None, tick=None):
        """ Once in a while queries will persist additional
            information on the server. This information is available
            through the following methods of this class:

                - experiment_types
                - assessor_types
                - scan_types
                - reconstruction_types

            It is also transparently used in insert operations.

            Parameters
            ----------
            auto: boolean
                True to enable auto learn. False to disable.
            tick: int
                Every 'tick' seconds, if a query is issued, additional
                information will be persisted.

            See Also
            --------
            :func:`EObject.insert`
        """
        if auto is not None:
            self._auto = auto
        if tick is not None:
            self._tick = tick

    def datatypes(self, pattern='*', fields_pattern=None):
        """ Discovers the datatypes and datafields of the database.

            Parameters
            ----------
            pattern: string
                Pattern for the datatype. May include wildcards.
            fields_pattern: string
                - Pattern for the datafields -- may include wildcards.
                - If specified, datafields will be returned instead of
                  datatypes.

            Returns
            -------
            list : datatypes or datafields depending on the argument usage.
        """
        self._intf._get_entry_point()

        search_els = self._get_json('%s/search/elements?format=json' %
                                    self._intf._get_entry_point())

        if not fields_pattern and ('*' in pattern or '?' in pattern):
            return get_column(search_els, 'ELEMENT_NAME', pattern)

        else:
            fields = []
            for datatype in get_column(search_els, 'ELEMENT_NAME', pattern):
                df = self._datafields(datatype, fields_pattern or '*', True)
                fields.extend(df)

            return fields

    def _datafields(self, datatype, pattern='*', prepend_type=True):
        self._intf._get_entry_point()

        search_fds = self._get_json('%s/search/elements/%s?format=json' %
                                    (self._intf._get_entry_point(),
                                        datatype))

        fields = get_column(search_fds, 'FIELD_ID', pattern)

        return ['%s/%s' % (datatype, field)
                if prepend_type else field
                for field in fields
                if '=' not in field and 'SHARINGSHAREPROJECT' not in field
                ]

    def experiment_types(self):
        """ Returns the datatypes used at the experiment level in this
            database.

            See Also
            --------
            :func:`Inspector.set_autolearn`
        """
        return self._resource_types('experiment')

    def assessor_types(self):
        """ Returns the datatypes used at the assessor level in this
            database.

            See Also
            --------
            :func:`Inspector.set_autolearn`
        """
        return self._resource_types('assessor')

    def reconstruction_types(self):
        """ Returns the datatypes used at the reconstruction level in this
            database.

            See Also
            --------
            :func:`Inspector.set_autolearn`
        """
        return self._resource_types('reconstruction')

    def scan_types(self):
        """ Returns the datatypes used at the scan level in this
            database.

            See Also
            --------
            :func:`Inspector.set_autolearn`
        """
        return self._resource_types('scan')

    def field_values(self, field_name):
        """ Look for the values a specific datafield takes in the database.
        """
        self._intf._get_entry_point()

        search_tbl = Search(field_name.split('/')[0],
                            [field_name], self._intf
                            )

        criteria = [('%s/ID' % field_name.split('/')[0], 'LIKE', '%'), 'AND']

        return list(set([val
                         for entry in search_tbl.where(criteria)
                         for val in entry.values()
                         ])
                    )

    def project_values(self):
        """ Look for the values a the project level in the database.

            .. note::
                Is equivalent to interface.select.projects().get()
        """
        self._intf._get_entry_point()

        return get_column(self._get_json(
                '%s/projects' % self._intf._entry), 'ID')

    def subject_values(self, project=None):
        """ Look for the values a the subject level in the database.

            .. note::
                Is equivalent to interface.select('//subjects').get()
        """
        self._intf._get_entry_point()

        uri = '%s/subjects?columns=ID' % self._intf._entry

        if project is not None:
            uri += '&project=%s' % project

        return get_column(self._get_json(uri), 'ID')

    def experiment_values(self, datatype, project=None):
        """ Look for the values a the experiment level for a given datatype
            in the database.

            .. note::
                The  datatype should be one of Inspector.experiment_types()

            Parameters
            ----------
            datatype: string
                An experiment type. eg: xnat:mrsessiondata
            project: string
                Optional. Restrict operation to a project.
        """
        self._intf._get_entry_point()

        uri = '%s/experiments?columns=ID' % self._intf._entry
        if datatype is not None:
            uri += '&xsiType=%s' % datatype
        if project is not None:
            uri += '&project=%s' % project

        return get_column(self._get_json(uri), 'ID')

    def assessor_values(self, experiment_type, project=None):
        """ Look for the values at the assessor level for a given experiment
            type in the database.

            .. note::
               The experiment type should be one of
               Inspector.experiment_types()

            .. warning::
                Depending on the number of elements the operation may
                take a while.

            Parameters
            ----------
            datatype: string
                An experiment type. eg: xnat:mrsessiondata
            project: string
                Optional. Restrict operation to a project.
        """
        return self._sub_experiment_values('assessor',
                                           project, experiment_type)

    def scan_values(self, experiment_type, project=None):
        """ Look for the values at the scan level for a given experiment
            type in the database.

            .. note::
               The experiment type should be one of
               Inspector.experiment_types()

            .. warning::
                Depending on the number of elements the operation may
                take a while.

            Parameters
            ----------
            datatype: string
                An experiment type.
            project: string
                Optional. Restrict operation to a project.
        """
        return self._sub_experiment_values('scan', project, experiment_type)

    def reconstruction_values(self, experiment_type, project=None):
        """ Look for the values at the reconstruction level for a given
            experiment type in the database.

            .. note::
               The experiment type should be one of
               Inspector.experiment_types()

            .. warning::
                Depending on the number of elements the operation may
                take a while.

            Parameters
            ----------
            datatype: string
                An experiment type.
            project: string
                Optional. Restrict operation to a project.
        """
        return self._sub_experiment_values('reconstruction',
                                           project, experiment_type)

    def structure(self):
        """ Displays the keywords structure used in XNAT REST API.
        """
        def traverse(coll, lvl):
            for key in schema.resources_tree[coll]:
                print('%s+%s' % (' ' * lvl, key.upper()))

                datatypes = set([
                        self._intf._struct[uri]
                        for uri in self._intf._struct.keys()
                        if uri.split('/')[-2] == key
                        ])

                if datatypes != set():
                    print('%s  %s' % (' ' * lvl, '-' * len(key)))

                for datatype in datatypes:
                    print('%s- %s' % (' ' * lvl, datatype))

                if key in schema.resources_tree.keys():
                    traverse(key, lvl + 4)

        print('- %s' % 'PROJECTS')
        traverse('projects', 4)

    def _sub_experiment_values(self, sub_exp, project, experiment_type):
        self._intf._get_entry_point()

        values = []

        column = '%s/%ss/%s/id' % \
            (experiment_type.lower(), sub_exp, sub_exp)

        sub_exps = '%s/experiments?columns=ID,%s' % (self._intf._entry,
                                                     column
                                                     )

        if project is not None:
            sub_exps += '&project=%s' % project

        values = get_column(self._get_json(sub_exps), column)

        return list(set(values))

    def _resource_struct(self, name):

        return self._intf._struct

    def _resource_types(self, name):
        return list(set(self._resource_struct(name).values()))


class GraphData(object):
    def __init__(self, interface):
        self._intf = interface
        self._struct = interface._struct

    # def link(self, subjects, fields):

    #     criteria = [('xnat:subjectData/SUBJECT_ID', '=', _id)
    #                 for _id in subjects
    #                 ]
    #     criteria += ['OR']
    #     # variables = ['xnat:subjectData/SUBJECT_ID'] + fields

    #     subject_id = 'xnat:subjectData/SUBJECT_ID'

    #     for field in fields:

    #         field_tbl = self._intf.select('xnat:subjectData',
    #                                       [subject_id, field]
    #                                       ).where(criteria)
    #         head = field_tbl.headers()
    #         head.remove('subject_id')
    #         head = head[0]
    #         possible = set(field_tbl.get(head))

    #         groups = {}

    #         for val in possible:
    #             groups[val] = field_tbl.where(**{head:val}
    #                                             ).select('subject_id')

    #     return groups

    def datatypes(self, pattern='*'):
        graph = nx.DiGraph()
        graph.add_node('datatypes')
        graph.labels = {'datatypes': 'datatypes'}
        graph.weights = {'datatypes': 100.0}

        datatypes = self._intf.inspect.datatypes(pattern)
        namespaces = set([dat.split(':')[0] for dat in datatypes])

        for ns in namespaces:
            graph.add_edge('datatypes', ns)
            graph.weights[ns] = 70.0

            for dat in datatypes:
                if dat.startswith(ns):
                    graph.add_edge(ns, dat)
                    graph.weights[dat] = 40.0

        return graph

    def rest_resource(self, name):
        resource_types = self._intf.inspect._resource_types(name)

        graph = nx.DiGraph()
        graph.add_node(name)
        graph.labels = {name: name}
        graph.weights = {name: 100.0}

        namespaces = set([exp.split(':')[0] for exp in resource_types])

        for ns in namespaces:
            graph.add_edge(name, ns)
            graph.weights[ns] = 70.0

            for exp in resource_types:
                if exp.startswith(ns):
                    graph.add_edge(ns, exp)
                    graph.weights[exp] = 40.0

        return graph

    def field_values(self, field_name):

        search_tbl = Search(field_name.split('/')[0],
                            [field_name], self._intf
                            )

        criteria = [('%s/ID' % field_name.split('/')[0], 'LIKE', '%'), 'AND']

        dist = {}

        for entry in search_tbl.where(criteria):
            for val in entry.values():
                dist.setdefault(val, 1.0)
                dist[val] += 1

        graph = nx.Graph()
        graph.add_node(field_name)
        graph.weights = dist
        graph.weights[field_name] = 100.0

        for val in dist.keys():
            graph.add_edge(field_name, val)

        return graph

    def architecture(self, with_datatypes=True):
        graph = nx.DiGraph()
        graph.add_node('projects')
        graph.labels = {'projects': 'projects'}
        graph.weights = {'projects': 100.0}

        def traverse(lkw, as_lkw):

            for key in schema.resources_tree[lkw]:
                as_key = '%s_%s' % (as_lkw, key)
                weight = (1 - len(as_key*2)/100.0) * 100
                graph.add_edge(as_lkw, as_key)
                graph.labels[as_key] = key
                graph.weights[as_key] = weight

                if with_datatypes:
                    for uri in self._struct.keys():
                        if uri.split('/')[-2] == key:
                            datatype = self._struct[uri]
                            graph.add_edge(as_key, datatype)
                            graph.weights[datatype] = 10
                            graph.labels[datatype] = datatype

                traverse(key, as_key)

        traverse('projects', 'projects')

        return graph


class PaintGraph(object):
    def __init__(self, interface):
        self._intf = interface
        self.get_graph = interface._get_graph

    def architecture(self, with_datatypes=True, save=None):
        graph = self.get_graph.architecture(with_datatypes)

        plt.figure(figsize=(8, 8))
        pos = graphviz_layout(graph, prog='twopi', args='')

        # node_size = [(float(graph.degree(v)) * 5)**3 for v in graph]
        # node_size = [graph.weights[v] ** 2 for v in graph]
        # node_color = [float(graph.degree(v)) for v in graph]
        # node_color = [graph.weights[v] ** 2 for v in graph]

        cost = lambda v: float(graph.degree(v)) ** 3 + \
            graph.weights[v] ** 2

        costs = norm_costs([cost(v) for v in graph], 10000)

        nx.draw(graph, pos, labels=graph.labels,
                node_size=costs, node_color=costs,
                font_size=13, font_color='orange',
                font_weight='bold', with_labels=True
                )

        plt.axis('off')

        if save is not None:
            plt.savefig(save)

        plt.show()

    def experiments(self, save=None):
        graph = self.get_graph.rest_resource('experiments')
        self._draw_rest_resource(graph, save=None)

    def assessors(self, save=None):
        graph = self.get_graph.rest_resource('assessors')
        self._draw_rest_resource(graph, save=None)

    def reconstructions(self, save=None):
        graph = self.get_graph.rest_resource('reconstructions')
        self._draw_rest_resource(graph, save=None)

    def scans(self):
        graph = self.get_graph.rest_resource('scans')
        self._draw_rest_resource(graph)

    def _draw_rest_resource(self, graph, save=None):
        plt.figure(figsize=(8, 8))
        pos = graphviz_layout(graph, prog='twopi', args='')

        cost = lambda v: float(graph.degree(v)) ** 3 + \
            graph.weights[v] ** 2

        node_size = [cost(v) for v in graph]
        # node_size = [graph.weights[v] ** 2 for v in graph]
        # node_color = [float(graph.degree(v)) for v in graph]
        node_color = [cost(v) for v in graph]

        nx.draw(graph, pos,
                node_size=node_size, node_color=node_color,
                font_size=13, font_color='green', font_weight='bold'
                )

        plt.axis('off')

        if save is not None:
            plt.savefig(save)

        plt.show()

    def datatypes(self, pattern='*', save=None):
        graph = self.get_graph.datatypes(pattern)

        plt.figure(figsize=(8, 8))
        pos = graphviz_layout(graph, prog='twopi', args='')

        cost = lambda v: float(graph.degree(v)) ** 3 + \
            graph.weights[v] ** 2

        node_size = [cost(v) for v in graph]
        # node_size = [graph.weights[v] ** 2 for v in graph]
        # node_color = [float(graph.degree(v)) for v in graph]
        node_color = [cost(v) for v in graph]

        nx.draw(graph, pos,
                node_size=node_size, node_color=node_color,
                font_size=13, font_color='green', font_weight='bold',
                with_labels=True
                )

        plt.axis('off')

        if save is not None:
            plt.savefig(save)

        plt.show()

    def field_values(self, field_name, save=None):
        graph = self.get_graph.field_values(field_name)

        plt.figure(figsize=(8, 8))
        pos = graphviz_layout(graph, prog='twopi', args='')

        cost = lambda v: graph.weights[v]

        graph.weights[field_name] = max([cost(v) for v in graph]) / 2.0

        costs = norm_costs([cost(v) for v in graph], 10000)

        nx.draw(graph, pos,
                node_size=costs, node_color=costs,
                font_size=13, font_color='black',
                font_weight='bold', with_labels=True
                )

        plt.axis('off')

        if save is not None:
            plt.savefig(save)

        plt.show()


def norm_costs(costs, norm=1000):
    max_cost = max(costs)

    return [(cost / max_cost) * norm for cost in costs]


# class GraphDrawer(object):
#     def __init__(self, interface):
#         self._intf = interface

#     def datatypes(self, project):
#         self._intf.connection.set_strategy('offline')

#         experiments_types = self._intf.inspect.datatypes.experiments(project)

#         labels = {project:project, 'Experiments':'Experiments'}

#         g = nx.Graph()

#         g.add_edge(project, 'Experiments', {'weight':1})

#         for exp_type in experiments_types:
#             g.add_edge('Experiments', exp_type, {'weight':8})
#             labels[exp_type] = exp_type

#         pos = nx.graphviz_layout(g, prog='twopi', args='')

#         nx.draw_networkx_nodes(g, pos,
#                                nodelist=[project],
#                                node_color='green', alpha=0.7,
#                                node_size=2500, node_shape='s')

#         nx.draw_networkx_nodes(g, pos,
#                                nodelist=['Experiments'],
#                                node_color='blue', alpha=0.7,
#                                node_size=2000, node_shape='p')

#         nx.draw_networkx_nodes(g, pos,
#                                nodelist=experiments_types,
#                                node_color='red', alpha=0.7,
#                                node_size=1500, node_shape='o')

#         nx.draw_networkx_edges(g, pos, width=2, alpha=0.5,
#                                edge_color='black')

#         nx.draw_networkx_labels(g, pos, labels,
#                                 alpha=0.9, font_size=8,
#                                 font_color='black', font_weight='bold')

#         plt.axis('off')
#         plt.show()

#         self._intf.connection.revert_strategy()

class SchemasInspector(object):
    def __init__(self, interface):
        self._intf = interface

    def __call__(self):
        self._intf.manage.schemas()

        for xsd in self._intf.manage.schemas():
            print('-'*40)
            print(xsd.upper())
            print('-'*40)
            print()

            trees = self._intf.manage.schemas._trees[xsd]
            for datatype in schema.datatypes(trees):
                print('[%s]' % datatype)
                print()

                for path in schema.datatype_attributes(trees, datatype):
                    print(path)

                print()

    def look_for(self, element_name, datatype_name=None):
        paths = []
        self._intf.manage.schemas._init()

        if ':' in element_name:
            for root in self._intf.manage.schemas._trees.values():
                paths.extend(schema.datatype_attributes(root, element_name))
            return paths

        for xsd in self._intf.manage.schemas():
            # nsmap = self._intf.manage.schemas._trees[xsd].nsmap
            trees = self._intf.manage.schemas._trees[xsd]
            if datatype_name is not None:
                datatypes = [datatype_name]
            else:
                datatypes = schema.datatypes(trees)

            for datatype in datatypes:
                for path in schema.datatype_attributes(trees, datatype):
                    if element_name in path:
                        paths.append(path)

        return paths



from uuid import uuid1
from . import skip_if_no_network
from pyxnat import Interface
import os.path as op


fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
central = Interface(config=fp)
project = central.select.project('nosetests5')

prov = {
    'program': 'young',
    'timestamp': '2011-03-01T12:01:01.897987',
    'user': 'angus',
    'machine': 'war',
    'platform': 'linux',
    }

sid = uuid1().hex
eid = uuid1().hex
aid = uuid1().hex


@skip_if_no_network
def test_provenance():
    assessor = project.subject(sid).experiment(eid).assessor(
        aid).insert(use_label=True)
    assert assessor.exists()
    assessor.provenance.set(prov)
    _prov = assessor.provenance.get()[0]

    assert prov['program'] == _prov['program'], \
        "Subject: %s Study: %s Prov: %s" % (sid, eid, aid)

# def test_del_provenance():
#     assessor.provenance.delete()
#     print assessor.provenance.get()
#     assert assessor.provenance.get()[0] == []


@skip_if_no_network
def test_provenance_cleanup():
    project.subject(sid).delete()
    assert not project.subject(sid).exists()



from pyxnat import xpass


def test_find_plus_line():
    print("Testing find_plus_line")
    test = ['hello', '+hello world', '']
    assert(xpass.find_plus_line(test) == 'hello world')
    test2 = ['', 'hello world']
    assert(xpass.find_plus_line(test2) is None)
    test3 = []
    assert(xpass.find_plus_line(test3) is None)


def test_find_token():
    print("Testing find_token")
    str = "hello,world"
    assert(xpass.find_token(',', str) == ('hello', 'world'))
    assert(xpass.find_token(' ', str) is None)


def test_parse_xnat_pass():
    print("Testing parse_xnat_pass")
    nothingLine = ""
    line = "+user@localhost:8080/xnat=password"
    lineWithSpaces = "+user@localhost:8080/xnat=password     "
    lineWithoutPlus = "user@localhost:8080/xnat=password"
    lineWithoutUser = "+@localhost:8080/xnat=password"
    lineWithoutHost = "+user=password"
    lineWithoutPass = "+user@localhost:8080/xnat"
    xp = xpass.parse_xnat_pass([nothingLine, line])
    assert(xp == {'p': 'password', 'host': 'localhost:8080/xnat', 'u': 'user'})
    xp = xpass.parse_xnat_pass([lineWithSpaces])
    assert(xp == {'p': 'password', 'host': 'localhost:8080/xnat', 'u': 'user'})
    assert(xpass.parse_xnat_pass([lineWithoutPlus]) is None)
    assert(xpass.parse_xnat_pass([lineWithoutUser]) is None)
    assert(xpass.parse_xnat_pass([lineWithoutHost]) is None)
    assert(xpass.parse_xnat_pass([lineWithoutPass]) is None)



import os
from uuid import uuid1
from pyxnat import Interface
import os.path as op
from . import skip_if_no_network


fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
central = Interface(config=fp)
project = central.select.project('nosetests5')

variables = {'Subjects': {'newgroup': {'foo': 'string', 'bar': 'int'}}}

sid = uuid1().hex
eid = uuid1().hex
cid = uuid1().hex

if not os.environ.get('PYXNAT_SKIP_NETWORK_TESTS'):
    sc = project.subject(sid).experiment(eid).scan(cid)
    scan = sc.insert(use_label=True)


@skip_if_no_network
def test_01_set_param():
    scan.set_param('foo', 'foostring')
    scan.set_param('bar', '1')

    assert scan.params() == ['foo', 'bar']


@skip_if_no_network
def test_02_get_params():
    assert scan.get_params() == ['foostring', '1']


@skip_if_no_network
def test_03_params_cleanup():
    project.subject(sid).delete()
    assert not project.subject(sid).exists()


@skip_if_no_network
def test_04_add_custom_variables():
    project.add_custom_variables(variables)


@skip_if_no_network
def test_05_get_custom_variables():
    project.get_custom_variables()



import os


def find_files(src):
    names = os.listdir(src)

    errors = []
    files = []

    for name in names:
        srcname = os.path.join(src, name)
        try:
            if os.path.islink(srcname):
                linkto = os.readlink(srcname)
                files.extend(find_files(linkto))
                # os.symlink(linkto, dstname)
            elif os.path.isdir(srcname):
                files.extend(find_files(srcname))
            else:
                files.append(srcname)
        except (IOError, os.error) as why:
            errors.append((srcname, str(why)))

    return files


def ensure_dir_exists(dir_path):
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    return os.path.isdir(dir_path)



XNAT_RESOURCE_NAMES = ['BBRC_VALIDATOR']


def tests(self, name, key=None):
    """
    Looks up a validation test based on the validator name and returns its
    result. If no test name is provided, then a dictionary is returned with
    all the tests from the validation report.

    Args:
        name (String): Root found in the name of the report
            (ex: FreeSurferValidator)
        key (String): Name of the test.
    Returns:
        A `pyxnat` File of the report.

    """
    import json
    j = [e for e in list(self.files('{}*.json'.format(name)))][0]
    j = json.loads(self._intf.get(j._uri).text)
    if key is None:
        return j
    else:
        return j[key]


def pdf(self, name):
    """
    Returns a validator report based on its name. Raises exception if
    multiple reports are found.

    Args:
        name (String): Root found in the name of the report
            (ex: FreeSurferValidator)
    Returns:
        A `pyxnat` File of the report.

    """
    files = list(self.files())
    pdf = {each._uri.split('/')[-1]: each for each in files
           if name in each._uri.split('/')[-1] and
           each._uri.split('/')[-1].endswith('.pdf')}

    if len(pdf.items()) == 0:
        raise Exception('No %s found' % name)

    keys = list(pdf.keys())
    if len(keys) != 1:
        raise Exception('Multiple matching reports found (%s)' % keys)

    f = pdf[list(pdf.keys())[0]]
    return f


def download_snapshot(self, name, fp):
    """
    Extract snapshot(s) from a validation report given its name.

    Args:
        name (String): Root found in the name of the report
            (ex: FreeSurferValidator)
        fp (String): Path to the file where the snapshot(s) will be saved.
    """

    import os
    import os.path as op
    import tempfile

    def extract_snapshots(fp):
        import fitz
        doc = fitz.open(fp)
        dn = op.dirname(fp)
        bn = op.splitext(op.basename(fp))[0]
        images = []
        xrefs = []
        for page in range(len(doc)):
            xrefs.extend([img[0] for img in doc.getPageImageList(page)])
        xrefs = sorted(set(xrefs))
        for k, xref in enumerate(xrefs[1:]):
            pix = fitz.Pixmap(doc, xref)
            fp_snap = op.join(dn, "{}_{}.png".format(bn, k))
            if pix.n >= 5:    # CMYK: convert to RGB first
                pix = fitz.Pixmap(fitz.csRGB, pix)
            pix.writePNG(fp_snap)
            images.append(fp_snap)
        return images

    pdf = self.pdf(name)

    f, fp1 = tempfile.mkstemp(suffix='.pdf')
    os.close(f)
    pdf.get(dest=fp1)

    snaps = extract_snapshots(fp1)

    if len(snaps) == 0:
        raise Exception('No snapshot found in report. %s %s' % (name, fp1))

    files = []  # Will store the snapshots in their final folder
    for i, each in enumerate(snaps):
        bn, _ = op.splitext(fp)
        _, ext = op.splitext(each)
        interm = ''
        if len(snaps) > 1:  # if multiple snapshots then add # in filename
            interm = '_%s' % i
        fp2 = '%s%s%s' % (bn, interm, ext)  # create destination filepath
        cmd = 'mv %s %s' % (each, fp2)
        os.system(cmd)
        files.append(fp2)

    return files



from uuid import uuid1
from pyxnat import Interface
from pyxnat import jsonutil
import os.path as op

fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
central = Interface(config=fp)
search_name = uuid1().hex
search_template_name = uuid1().hex


def test_datatypes():
    assert 'xnat:subjectData' in central.inspect.datatypes()


def test_datafields():
    assert 'xnat:subjectData/DOB' in \
                    central.inspect.datatypes('xnat:subjectData')
    assert 'xnat:subjectData/DOB' in \
           central.inspect.datatypes('xnat:subjectData', '*')


def test_fieldvalues():
    assert len(central.inspect.field_values('xnat:subjectData/SUBJECT_ID')
               ) != 0


def test_inspect_resources():
    assert 'OAS1_0440_MR1' in \
        central.inspect.experiment_values('xnat:mrSessionData',
                                          'CENTRAL_OASIS_CS')

    assert 'OAS1_0286_MR1_FSEG' in \
        central.inspect.assessor_values('xnat:mrSessionData',
                                        'CENTRAL_OASIS_CS')

    assert 'mpr-1' in \
        central.inspect.scan_values('xnat:mrSessionData',
                                    'CENTRAL_OASIS_CS')

    # just coverage
    assert isinstance(central.inspect.experiment_types(), list)
    assert isinstance(central.inspect.assessor_types(), list)
    assert isinstance(central.inspect.scan_types(), list)
    assert isinstance(central.inspect.reconstruction_types(), list)
    assert isinstance(central.inspect.project_values(), list)
    assert isinstance(central.inspect.subject_values(), list)


def test_search():
    results = central.select(
        'xnat:mrSessionData',
        central.inspect.datatypes('xnat:mrSessionData')
        ).where([('xnat:mrSessionData/SCANNER', 'LIKE', '*GE*'), 'AND'])

    assert isinstance(results, jsonutil.JsonTable)


def test_save_search():
    central.manage.search.save(
        search_name, 'xnat:mrSessionData',
        central.inspect.datatypes('xnat:mrSessionData'),
        [('xnat:mrSessionData/SCANNER', 'LIKE', '*GE*'), 'AND'])

    assert search_name in central.manage.search.saved()


def test_get_search():
    results = central.manage.search.get(search_name)
    assert isinstance(results, jsonutil.JsonTable)


def test_delete_search():
    central.manage.search.delete(search_name)
    assert search_name not in central.manage.search.saved()


def test_save_search_template():
    central.manage.search.save_template(
        search_template_name, 'xnat:mrSessionData',
        central.inspect.datatypes('xnat:mrSessionData'),
        [('xnat:mrSessionData/SCANNER', 'LIKE', '*GE*'), 'AND']
        )

    assert search_template_name in central.manage.search.saved_templates()


def test_delete_search_template():
    central.manage.search.delete_template(search_template_name)
    assert search_template_name not in \
        central.manage.search.saved_templates()



"""
PyXNAT
======

:mod:`pyxnat` provides an API to access data on XNAT servers
(see http://xnat.org).

Visit https://pyxnat.github.io/pyxnat for more information.
"""

from .version import VERSION as __version__

from .core import Interface
from .core import SearchManager
from .core import Select
from .core import Inspector
from .core import Users
from .core import attributes
from .core import help
from .core import interfaces
from .core import resources
from .core import manage
from .core import schema
from .core import select
from .core import users
from .core import jsonutil
from .core import uriutil
from .core import xpass
from .core import xpath_store



XNAT_RESOURCE_NAME = 'ASHS'


def volumes(self, mode='corr_nogray'):
    import pandas as pd

    f = list(self.files('*icv.txt'))[0]
    uri = f._uri
    res = self._intf.get(uri).text.split(' ')

    f = list(self.files('*_left_%s_volumes.txt' % mode))[0]
    uri = f._uri
    resl = self._intf.get(uri).text.split('\n')

    f = list(self.files('*_right_%s_volumes.txt' % mode))[0]
    uri = f._uri
    resr = self._intf.get(uri).text.split('\n')

    table = []
    for resx in [resl, resr]:
        for line in resx:
            if line == '':
                continue
            line = line.split(' ')
            s = line[0]
            side = line[1]
            region = line[2]
            i = int(line[-2])
            m = float(line[-1])
            table.append([s, side, region, i, m])

    table.append([res[0], None, 'tiv', None, float(res[1].rstrip('\n'))])

    columns = ['subject', 'side', 'region', 'n_slices', 'volume']
    return pd.DataFrame(table, columns=columns)



#!/usr/bin/env python

import setuptools
import os.path as op


def get_version():
    try:
        import re
        import os.path as op
        fp = op.join(op.dirname(__file__), 'pyxnat', 'version.py')
        s = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format('VERSION'),
                      open(fp).read())
        return s.group(1)
    except Exception:
        raise RuntimeError("No version found")


this_directory = op.abspath(op.dirname(__file__))
with open(op.join(this_directory, 'README.rst')) as f:
    long_description = f.read()


setuptools.setup(name='bbrc-pyxnat',
                 version=get_version(),
                 summary='XNAT in Python',
                 author='Yannick Schwartz',
                 author_email='yannick.schwartz@cea.fr',
                 url='https://github.com/pyxnat/pyxnat',
                 packages=setuptools.find_packages(exclude=('doc*', 'tests')),
                 description='XNAT in Python',
                 long_description=long_description,
                 long_description_content_type='text/x-rst',
                 license='BSD',
                 classifiers=[
                      'Development Status :: 4 - Beta',
                      'Environment :: Console',
                      'Intended Audience :: Developers',
                      'Intended Audience :: Science/Research',
                      'Intended Audience :: Education',
                      'License :: OSI Approved :: BSD License',
                      'Operating System :: OS Independent',
                      'Topic :: Scientific/Engineering',
                      'Topic :: Utilities',
                      'Topic :: Internet :: WWW/HTTP',
                      'Programming Language :: Python :: 3.6',
                      'Programming Language :: Python :: 3.7',
                      'Programming Language :: Python :: 2.7',
                 ],

                 platforms='any',
                 scripts=['bin/sessionmirror.py'],
                 install_requires=['lxml>=4.3',
                                   'requests>=2.20',
                                   'pathlib>=1.0',
                                   'six>=1.15',
                                   'future>=0.16'],
                 package_data={'pyxnat': ['README.rst'], },

                 )



from .jsonutil import JsonTable


class ArrayData(object):

    def __init__(self, interface):
        self._intf = interface

    def _get_array(self, query_string, project_id=None,
                   subject_id=None, subject_label=None,
                   experiment_id=None, experiment_label=None,
                   experiment_type='xnat:subjectAssessorData',
                   columns=None, constraints=None
                   ):

        if constraints is None:
            constraints = {}

        uri = '%s/experiments?xsiType=%s' % (self._intf._get_entry_point(),
                                             experiment_type)

        if project_id is not None:
            uri += '&project=%s' % project_id

        if subject_id is not None:
            uri += '&%s/subject_id=%s' % (experiment_type, subject_id)

        # Subject Label is only held in the xnat:subjectData so look for it
        # there. This should join against whatever the experiment type is.
        if subject_label is not None:
            uri += '&xnat:subjectData/label=%s' % (subject_label)

        if experiment_id is not None:
            uri += '&ID=%s' % experiment_id

        if experiment_label is not None:
            uri += '&label=%s' % experiment_label

        uri += query_string

        if constraints != {}:
            uri += ',' + ','.join(constraints.keys())

        if columns is not None:
            uri += ',' + ','.join(columns)

        c = {}

        [c.setdefault(key.lower(), value)
         for key, value in constraints.items()
         ]

        return JsonTable(self._intf._get_json(uri)).where(**c)

    def experiments(self, project_id=None, subject_id=None, subject_label=None,
                    experiment_id=None, experiment_label=None,
                    experiment_type='xnat:subjectAssessorData',
                    columns=None,
                    constraints=None):

        """ Returns a list of all visible experiment IDs of the specified
            type, filtered by optional constraints.

            Parameters
            ----------
            project_id: string
                Name pattern to filter by project ID.
            subject_id: string
                Name pattern to filter by subject ID.
            subject_label: string
                Name pattern to filter by subject ID.
            experiment_id: string
                Name pattern to filter by experiment ID.
            experiment_label: string
                Name pattern to filter by experiment ID.
            experiment_type: string
                xsi path type; e.g. 'xnat:mrSessionData'
            columns: list
                Values to return.
            constraints: dict
                Dictionary of xsi_type (key--) and parameter (--value)
                pairs by which to filter.
            """

        query_string = '&columns=ID,project,%s/subject_id' % experiment_type

        return self._get_array(query_string, project_id,
                               subject_id, subject_label,
                               experiment_id, experiment_label,
                               experiment_type, columns, constraints
                               )

    def mrsessions(self, project_id=None, subject_id=None, subject_label=None,
                   experiment_id=None, experiment_label=None,
                   columns=None,
                   constraints=None):

        """ Returns a list of all MR sessions, filtered by optional constraints.

            Parameters
            ----------
            project_id: string
                Name pattern to filter by project ID.
            subject_id: string
                Name pattern to filter by subject ID.
            subject_label: string
                Name pattern to filter by subject ID.
            experiment_id: string
                Name pattern to filter by experiment ID.
            experiment_label: string
                Name pattern to filter by experiment ID.
            columns: list
                Values to return.
            constraints: dict
                Dictionary of xsi_type (key--) and parameter (--value)
                pairs by which to filter.
            """

        return self.experiments(project_id, subject_id, subject_label,
                                experiment_id, experiment_label,
                                'xnat:mrSessionData', columns, constraints
                                )

    def scans(self, project_id=None, subject_id=None, subject_label=None,
              experiment_id=None, experiment_label=None,
              experiment_type='xnat:imageSessionData',
              scan_type='xnat:imageScanData',
              columns=None,
              constraints=None
              ):

        """ Returns a list of all visible scan IDs of the specified type,
            filtered by optional constraints.

            Parameters
            ----------
            project_id: string
                Name pattern to filter by project ID.
            subject_id: string
                Name pattern to filter by subject ID.
            subject_label: string
                Name pattern to filter by subject ID.
            experiment_id: string
                Name pattern to filter by experiment ID.
            experiment_label: string
                Name pattern to filter by experiment ID.
            experiment_type: string
                xsi path type; e.g. 'xnat:mrSessionData'
            scan_type: string
                xsi path type; e.g. 'xnat:mrScanData', etc.
            columns: list
                Values to return.
            constraints: dict
                Dictionary of xsi_type (key--) and parameter (--value)
                pairs by which to filter.
            """

        query_string = '&columns=ID,project,%s/subject_id,%s/ID' % (
            experiment_type, scan_type)

        array = self._get_array(query_string, project_id,
                                subject_id, subject_label,
                                experiment_id, experiment_label,
                                experiment_type, columns, constraints)
        id_key = ('%s/ID' % scan_type).lower()
        return JsonTable([i for i in array if i[id_key]])

    def mrscans(self, project_id=None, subject_id=None, subject_label=None,
                experiment_id=None, experiment_label=None,
                columns=None,
                constraints=None):

        """ Returns a list of all MR scans, filtered by optional constraints.

            Parameters
            ----------
            project_id: string
                Name pattern to filter by project ID.
            subject_id: string
                Name pattern to filter by subject ID.
            subject_label: string
                Name pattern to filter by subject ID.
            experiment_id: string
                Name pattern to filter by experiment ID.
            experiment_label: string
                Name pattern to filter by experiment ID.
            columns: list
                Values to return.
            constraints: dict
                Dictionary of xsi_type (key--) and parameter (--value)
                pairs by which to filter.
            """

        return self.scans(project_id, subject_id, subject_label,
                          experiment_id, experiment_label,
                          'xnat:mrSessionData', 'xnat:mrScanData',
                          columns, constraints
                          )

    def search_experiments(self,
                           project_id=None,
                           subject_id=None,
                           subject_label=None,
                           experiment_type='xnat:subjectAssessorData',
                           columns=None,
                           constraints=None
                           ):

        """ Returns a list of all visible experiment IDs of the
            specified type, filtered by optional constraints. This
            function is a shortcut using the search engine.

            Parameters
            ----------
            project_id: string
                Name pattern to filter by project ID.
            subject_id: string
                Name pattern to filter by subject ID.
            subject_label: string
                Name pattern to filter by subject ID.
            experiment_type: string
                xsi path type must be a leaf session type.
                defaults to 'xnat:mrSessionData'
            columns: List[string]
                list of xsi paths for names of columns to return.
            constraints: list[(tupple)]
                List of tupples for comparison in the form (key, comparison,
                value) valid comparisons are: =, <, <=,>,>=, LIKE
            """

        if columns is None:
            columns = []

        where_clause = []

        if project_id is not None:
            item = ('%s/project' % experiment_type, "=", project_id)
            where_clause.append(item)
        if subject_id is not None:
            where_clause.append(('xnat:subjectData/ID', "=", subject_id))
        if subject_label is not None:
            where_clause.append(('xnat:subjectData/LABEL', "=", subject_label))

        if constraints is not None:
            where_clause.extend(constraints)

        if where_clause != []:
            where_clause.append('AND')

        if where_clause != []:
            sel = self._intf.select(experiment_type, columns=columns)
            table = sel.where(where_clause)
            return table
        else:
            table = self._intf.select(experiment_type, columns=columns)
            return table.all()



from pyxnat import Interface
import os.path as op
from . import skip_if_no_network

_modulepath = op.dirname(op.abspath(__file__))

fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
print(fp)
central = Interface(config=fp)

proj_1 = central.select.project('surfmask_smpl2')
subj_1 = proj_1.subject('CENTRAL05_S01120')
exp_1 = subj_1.experiment('CENTRAL05_E02681')
scan_1 = exp_1.scan('11')
resource_1 = exp_1.resource('obscure_algorithm_output')

proj_2 = central.select.project('NFB')
subj_2 = proj_2.subject('BOA')
exp_2 = subj_2.experiment('GHJ')
scan_2 = exp_2.scan('JKL')
resource_2 = scan_2.resource('IOP')


@skip_if_no_network
def test_project_exists():
    if proj_1.exists():
        assert isinstance(proj_1, object)
        assert str(proj_1) != '<Project Object> NFB'


@skip_if_no_network
def test_project_not_exists():
    if not proj_2.exists():
        assert isinstance(proj_2, object)
        assert str(proj_2) == '<Project Object> NFB'


@skip_if_no_network
def test_info_project():
    assert isinstance(proj_1, object)
    expected_output = '<Project Object> surfmask_smpl2 `Surface masking '\
        'samples 2` (private) 1 subject 1 MR experiment (owner: nosetests) '\
        '(created on 2020-10-22 15:23:39.458) https://central.xnat.org/data/'\
        'projects/surfmask_smpl2?format=html'
    assert list(sorted(str(proj_1))) == list(sorted(expected_output))


@skip_if_no_network
def test_subject_exists():
    if subj_1.exists():
        assert isinstance(subj_1, object)
        assert str(subj_1) != '<Subject Object> BOA'


@skip_if_no_network
def test_subject_not_exists():
    if not subj_2.exists():
        assert isinstance(subj_2, object)
        assert str(subj_2) == '<Subject Object> BOA'


@skip_if_no_network
def test_info_subject():
    assert isinstance(subj_1, object)
    expected_output = '<Subject Object> CENTRAL05_S01120 `001` (project: '\
        'surfmask_smpl2) (Gender: U) 1 experiment https://central.xnat.org/'\
        'data/projects/surfmask_smpl2/subjects/CENTRAL05_S01120?format=html'
    assert list(sorted(str(subj_1))) == list(sorted(expected_output))


@skip_if_no_network
def test_experiment_exists():
    if exp_1.exists():
        assert isinstance(exp_1, object)
        assert str(exp_1) != '<Experiment Object> GHJ'


@skip_if_no_network
def test_experiment_not_exists():
    if not exp_2.exists():
        assert isinstance(exp_2, object)
        assert str(exp_2) == '<Experiment Object> GHJ'


@skip_if_no_network
def test_info_experiment():
    assert isinstance(exp_1, object)
    expected_output = '<Experiment Object> CENTRAL05_E02681 `001_obscured` (subject: '\
        'CENTRAL05_S01120 `001`) (project: surfmask_smpl2) 4 scans 1 resource '\
        '(created on 2020-10-22 15:24:30.139) https://central.xnat.org/'\
        'data/projects/surfmask_smpl2/subjects/CENTRAL05_S01120/experiments/'\
        'CENTRAL05_E02681?format=html'
    assert list(sorted(str(exp_1))) == list(sorted(expected_output))


@skip_if_no_network
def test_scan_exists():
    if scan_1.exists():
        assert isinstance(scan_1, object)
        assert str(scan_1) != '<Scan Object> JKL'


@skip_if_no_network
def test_scan_not_exists():
    if not scan_2.exists():
        assert isinstance(scan_2, object)
        assert str(scan_2) == '<Scan Object> JKL'


@skip_if_no_network
def test_info_scan():
    assert isinstance(scan_1, object)
    expected_output = '<Scan Object> 11 (`SPGR` 175 frames)  '\
        'https://central.xnat.org/data/projects/surfmask_smpl2/subjects/'\
        'CENTRAL05_S01120/experiments/CENTRAL05_E02681/scans/11?format=html'
    assert list(sorted(str(scan_1))) == list(sorted(expected_output))


@skip_if_no_network
def test_resource_exists():
    if resource_1.exists():
        assert isinstance(resource_1, object)
        assert str(resource_1) != '<Resource Object> IOP'


@skip_if_no_network
def test_resource_not_exists():
    if not resource_2.exists():
        assert isinstance(resource_2, object)
        assert str(resource_2) == '<Resource Object> IOP'


@skip_if_no_network
def test_info_resource():
    assert isinstance(resource_1, object)
    expected_output = '<Resource Object> 123361501 '\
        '`obscure_algorithm_output` (66 files 2.06 GB)'
    assert list(sorted(str(resource_1))) == list(sorted(expected_output))


@skip_if_no_network
def test_create_delete_create():
    p = central.select.project('nosetests5')
    from uuid import uuid1
    sid = uuid1().hex
    s = p.subject(sid)
    s.create()
    assert(s.exists())
    s.delete()
    s.create()
    s.delete()
    assert(not s.exists())



# Since pyxnat v1.2.1.0.post3, we introduced a mechanism to tune these objects
# and allow adding custom functions for specific types of resources.
#
# For instance, one could now directly write:
#
# resource = experiment.resource('FREESURFER6')
# aparc = resource.aparc()
# aseg = resource.aseg()
# and thus get access to FreeSurfer measurements.
#
# Another example, with the ASHS pipeline for hippocampal subfield
# segmentation:
#
# resource = experiment.resource('ASHS')
# volumes = resource.volumes()
#
# Again, this would only work provided that corresponding resources respect a
# certain naming and structure on XNAT. Here in this present example,
# FreeSurfer results are stored in resources called FREESURFER6 and the whole
# FreeSurfer folder (named after the subject) is stored in the resource.
# Having access to such additional functions would be conditioned by the
# existence of these resources with proper matching structure.
#
# Nevertheless, this mechanism has been implemented so as to get easily adapted
# to local configurations, by editing/adding this very same folder.
#
# Adding a custom function can be done simply as follows.
#
# In this same folder (pyxnat/core/derivatives/), edit any existing file or add
# a new one (filename does not matter):
#
# Define XNAT_RESOURCE_NAME. This variable names the XNAT resource which needs
# a custom function.
# Write the custom function with self as first parameter (self will be the
# pyxnat Resource object).
# If the resources which the custom function should be added has multiple names
# instead of a single one, their list may be provided under the variable
# XNAT_RESOURCE_NAMES.
#
# An example is provided in pyxnat/core/derivatives/ashs.py



XNAT_RESOURCE_NAMES = ['FREESURFER6', 'FREESURFER6_HIRES']


def hippoSfVolumes(self, mode='T1'):
    ''' Returns hippocampal subfield volumetry as estimated by FreeSurfer
    `recon-all`.'''
    import pandas as pd

    table = []
    files = list(self.files('*h.hippoSfVolumes-%s.v10.txt' % mode))
    for f in files:
        uri = f._uri

        res = self._intf.get(uri).text.split('\n')
        d1 = dict([each.split(' ') for each in res[:-1]])
        d2 = dict([('%s' % k, float(v))
                  for k, v in d1.items()])

        side = {'l': 'left', 'r': 'right'}[uri.split('/')[-1][0]]
        for region, value in d2.items():
            row = [side, region, value]
            table.append(row)

    columns = ['side', 'region', 'value']
    return pd.DataFrame(table, columns=columns)


def aparc(self, atlas='desikan-killiany'):
    ''' Returns cortical features as estimated by FreeSurfer `recon-all`.'''

    import pandas as pd

    volumes = ['BrainSegVol', 'BrainSegVolNotVent', 'BrainSegVolNotVentSurf',
               'CortexVol', 'SupraTentorialVol', 'SupraTentorialVolNotVent',
               'eTIV']

    unitless = ['NumVert']

    surfaces = ['WhiteSurfArea']

    thickness = ['MeanThickness']

    columns = ['StructName', 'NumVert', 'SurfArea', 'GrayVol', 'ThickAvg',
               'ThickStd', 'MeanCurv', 'GausCurv', 'FoldInd', 'CurvInd']

    table = []
    if atlas == 'desikan-killiany':
        ft = 'aparc'
    elif atlas == 'destrieux':
        ft = 'aparc.a2009s'
    else:
        raise Exception('Invalid atlas "%s". Should be either '
                        'desikan-killiany or destrieux.' % atlas)

    files = list(self.files('*h.%s.stats' % ft))
    for f in files:

        uri = f._uri
        side = {'l': 'left', 'r': 'right'}[uri.split('/')[-1][0]]

        res = self._intf.get(uri).text.split('\n')

        res2 = [e for e in res if e.startswith('# Measure ')]
        for each in volumes:
            m = float([e for e in res2
                       if each in e.split(', ')[1]][0].split(', ')[-2])
            table.append([None, 'Volume_mm3', each, m])

        for each in thickness:
            m = float([e for e in res2
                       if each in e.split(', ')[1]][0].split(', ')[-2])
            table.append([None, 'Thickness_mm', each, m])

        for each in surfaces:
            m = float([e for e in res2
                       if each in e.split(', ')[1]][0].split(', ')[-2])
            table.append([None, 'SurfArea_mm2', each, m])

        for each in unitless:
            m = float([e for e in res2
                       if each in e.split(', ')[1]][0].split(', ')[-2])
            table.append([None, 'unitless', each, m])

        res2 = [e for e in res if not e.startswith('#')]
        d2 = [[each for each in e.split(' ') if each != ''] for e in res2]
        d2 = pd.DataFrame(d2, columns=columns).dropna()
        d3 = d2.set_index('StructName').to_dict()

        for m, aparc in d3.items():
            for region, value in aparc.items():
                row = [side, m, region, value]
                table.append(row)

    columns = ['side', 'measurement', 'region', 'value']
    return pd.DataFrame(table, columns=columns)


def aseg(self):
    import pandas as pd
    columns = ['Index', 'SegId', 'NVoxels', 'Volume_mm3', 'StructName',
               'normMean', 'normStdDev', 'normMin', 'normMax', 'normRange']

    f = list(self.files('*aseg.stats'))[0]

    uri = f._uri

    res = self._intf.get(uri).text.split('\n')

    volumes = ['BrainSegVol', 'BrainSegVolNotVent', 'BrainSegVolNotVentSurf',
               'VentricleChoroidVol', 'lhCortexVol', 'rhCortexVol',  # 'Cortex'
               'lhCerebralWhiteMatterVol', 'rhCerebralWhiteMatterVol',
               'CerebralWhiteMatterVol', 'SubCortGrayVol', 'TotalGrayVol',
               'SupraTentorialVol', 'SupraTentorialVolNotVent',
               'SupraTentorialVolNotVentVox', 'MaskVol', 'eTIV']

    unitless = ['BrainSegVol-to-eTIV', 'MaskVol-to-eTIV', 'lhSurfaceHoles',
                'rhSurfaceHoles', 'SurfaceHoles']

    table = []
    res2 = [e for e in res if e.startswith('# Measure ')]
    for each in volumes:
        # [e for e in res2 if each == e.split(', ')[1]]
        m = float([e for e in res2
                   if each == e.split(', ')[1]][0].split(', ')[-2])
        table.append(['Volume_mm3', each, m])

    for each in unitless:
        m = float([e for e in res2
                   if each == e.split(', ')[1]][0].split(', ')[-2])
        table.append(['unitless', each, m])

    res2 = [e for e in res if not e.startswith('#')]
    d2 = [[each for each in e.split(' ') if each != ''] for e in res2]
    d2 = pd.DataFrame(d2, columns=columns).dropna()
    d3 = d2.set_index('StructName').to_dict()

    for m, aseg in d3.items():
        for region, value in aseg.items():
            row = [m, region, value]
            table.append(row)

    columns = ['measurement', 'region', 'value']
    return pd.DataFrame(table, columns=columns)


def normMean(self):

    from roistats import collect
    import pandas as pd
    f1 = list(self.files('*mri/aparc.a2009s+aseg.mgz'))[0]
    f2 = list(self.files('*mri/norm.mgz'))[0]
    s_label = f2._uri.split('/')[-3]
    e_id = f2._uri.split('/')[-7]
    f1.get('/tmp/aparc.mgz')
    f2.get('/tmp/T1.mgz')
    import json
    lut = dict([(int(e), v)
               for e, v in json.load(open('/tmp/d.json')).items()])
    roi_values = collect.roistats_from_maps(['/tmp/norm.mgz'],
                                            '/tmp/aparc.mgz',
                                            subjects=[(e_id, s_label)])
    df = pd.DataFrame(roi_values, columns=lut)
    return df.rename(columns=lut)
